
/******************************************************************************
* MODULE     : analytic.hpp
* DESCRIPTION: Analytic functions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_ANALYTIC_HPP
#define __MML_ANALYTIC_HPP
#include "complex.hpp"
#include "series.hpp"
namespace mml {
#define TMPL_DEF template<typename V>
#define TMPL template<typename V>
#define Analytic analytic<V>
#define Analytic_rep analytic_rep<V>
#define Positive typename V::_positive
#define Complex typename V::_complex
#define Series typename V::_series
TMPL class analytic;

/******************************************************************************
* Different variants of analytic functions
******************************************************************************/

class double_analytic {
public:
  typedef double _positive;
  typedef complexify<double> _complex;
  typedef series<_complex> _series;
  static inline _positive _plus_infinity () { return HUGE_VAL; }
  static inline _positive _abs (const _complex& z) { return mml::abs (z); }
  static inline _complex _construct (const _positive& x) { return x; }
};

inline bool is_infinity (double x) { return x >= HUGE_VAL; }

class exact_analytic {
public:
  typedef interval _positive;
  typedef complex _complex;
  typedef series<_complex> _series;
  static inline _positive _plus_infinity () {
    return _positive::plus_infinity; }
  static _positive _abs (const _complex& z) {
    interval x= expand (Re (z), DEFAULT_LOW);
    interval y= expand (Im (z), DEFAULT_LOW);
    return hypot (x, y); }
  static inline _complex _construct (const _positive& x) {
    return complex (as_real (center (x))); }
};

/******************************************************************************
* The analytic function class
******************************************************************************/

TMPL_DEF
class analytic_rep REP_STRUCT {
public:
  Series s;

protected:
  inline analytic_rep (const Series& s2): s (s2) {}
  inline virtual ~analytic_rep () {}

public:
  virtual generic expression (const generic& x) const = 0;
  virtual Positive radius (nat prec= 1) const = 0;
  virtual Positive bound (const Positive& r, nat prec= 1) const = 0;
  virtual Analytic move (const Complex& z) const = 0;
  virtual Complex eval (const Complex& z) const;
  Positive root_free_disk (nat prec= 1) const;
  Positive lower_bound (const Positive& r, nat prec= 1) const;
  Analytic continuate (const Complex& z) const;
  Complex evaluate (const Complex& z) const;
  friend class Analytic;
};

TMPL_DEF
class analytic {
INDIRECT_PROTO_1 (analytic, analytic_rep, V)
public:
  analytic (int c= 0);
  analytic (const Complex& c);
  analytic (const polynomial<Complex>& P);
  Complex operator [] (nat i) const;
  Complex operator () (const Complex& z) const;
};
INDIRECT_IMPL_1 (analytic, analytic_rep, typename V, V)

TMPL generic var (const Analytic& f) { (void) f; return the_series_var (); }

TMPL Analytic operator + (const Analytic& f, const Analytic& g);
TMPL Analytic operator - (const Analytic& f);
TMPL Analytic operator - (const Analytic& f, const Analytic& g);
TMPL Analytic operator * (const Complex& c, const Analytic& f);
TMPL Analytic operator * (const Analytic& f, const Complex& c);
TMPL Analytic operator / (const Analytic& f, const Complex& c);
TMPL Analytic operator % (const Analytic& f, const Complex& c);
TMPL Analytic operator * (const Analytic& f, const Analytic& g);
TMPL Analytic operator / (const Analytic& f, const Analytic& g);
TMPL Analytic operator % (const Analytic& f, const Analytic& g);

/******************************************************************************
* Helper functions
******************************************************************************/

template<typename Real> Real
smaller_radius (const Real& rr, nat prec, nat i) {
  Real r= min (rr, Real (1 << prec));
  Real a= Real (prec) / Real (prec + 1);
  return r * pow (a, i);
}

template<typename Real> Real
radius_between (const Real& r, const Real& R) {
  if (is_infinity (R)) return Real (2) * r;
  else return (r + R) / Real (2);
}

/******************************************************************************
* Default routines for analytic functions
******************************************************************************/

TMPL Complex
Analytic_rep::eval (const Complex& z) const {
  return move (z) [0];
}

TMPL Complex
Analytic::operator [] (nat n) const {
  return rep->s [n];
}

TMPL Complex
Analytic::operator () (const Complex& z) const {
  return rep->eval (z);
}

TMPL Positive
Analytic_rep::root_free_disk (nat prec) const {
  Positive C = V::_abs (s[0]);
  Positive R0= radius (prec);
  Positive R= R0, M= V::_plus_infinity ();
  for (nat i=1; Positive(2)*C < M; i++) {
    R= smaller_radius (R0, prec, i);
    M= bound (R);
  }
  return R * C / (M + C);
}

TMPL Positive
Analytic_rep::lower_bound (const Positive& r, nat prec) const {
  Positive C = V::_abs (s[0]);
  Positive R0= radius (prec);
  Positive R= R0, M= V::_plus_infinity ();
  for (nat i=1; Positive(2)*C < M; i++) {
    R= smaller_radius (R0, prec, i);
    M= bound (R);
  }
  return C - r * M / (R - r);
}

TMPL Analytic
Analytic_rep::continuate (const Complex& z) const {
  Positive R= radius ();
  Positive Z= V::_abs (z);
  if (Z < R) return move (z);
  else {
    Complex u= V::_construct (R / (Positive (2) * Z)) * z;
    //cout << "{" << z-u << "}"; cout.flush ();
    return move (u) -> continuate (z-u);
  }
}

/******************************************************************************
* Evaluation of analytic functions
******************************************************************************/

TMPL Complex
Analytic_rep::evaluate (const Complex& z) const {
  assert (false); // only implemented for specific variants
}

template<> double_analytic::_complex
analytic_rep<double_analytic>::evaluate
(const double_analytic::_complex& z) const;

template<> exact_analytic::_complex
analytic_rep<exact_analytic>::evaluate
(const exact_analytic::_complex& z) const;

/******************************************************************************
* Hashing and printing analytic functions
******************************************************************************/

TMPL nat
hash (const Analytic& s) {
  register nat i, h= 7531;
  for (i=0; i<4; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash (s[i]);
  return h;
}

TMPL generic
as_generic (const Analytic& f) {
  if (the_series_expr) return f->expression (var (f));
  generic s= 0;
  for (nat i=0; i<the_series_prec; i++)
    s= plus (s, times (as_generic (f[i]), power (var (f), generic (i))));
  s= plus (s, generic ("O") (power (var (f), generic (the_series_prec))));
  return s;
}

PRINT_SUGAR(TMPL,Analytic)

/******************************************************************************
* Explicit analytic functions
******************************************************************************/

TMPL_DEF
class polynomial_analytic_rep: public Analytic_rep {
  polynomial<Complex> P;
public:
  polynomial_analytic_rep (const polynomial<Complex>& P2):
    analytic_rep<V> (P2), P (P2) {}
  generic expression (const generic& x) const {
    (void) x;
    return as_generic (this->s); }
  Positive radius (nat prec) const {
    (void) prec;
    return V::_plus_infinity (); }
  Positive bound (const Positive& r, nat prec) const {
    (void) prec;
    Positive sum (V::_abs (this->s[0])), prod (r);
    for (nat i=1; i<dim(P); i++) {
      sum += V::_abs (this->s[i]) * prod;
      prod *= r;
    }
    return sum;
  }
  Analytic move (const Complex& z) const {
    polynomial<Complex> Q= seq (z, Complex (1));
    return Analytic (compose (P, Q));
  }
};

TMPL
Analytic::analytic (int c) {
  rep= new polynomial_analytic_rep<V> (polynomial<Complex> (Complex (c), 0));
}

TMPL
Analytic::analytic (const Complex& c) {
  rep= new polynomial_analytic_rep<V> (polynomial<Complex> (c, 0));
}

TMPL
Analytic::analytic (const polynomial<Complex>& P) {
  rep= new polynomial_analytic_rep<V> (P);
}

/******************************************************************************
* Sums and differences
******************************************************************************/

TMPL_DEF
class sum_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  sum_analytic_rep (const Analytic& f2, const Analytic& g2):
    analytic_rep<V> (f2->s + g2->s), f(f2), g(g2) {}
  generic expression (const generic& x) const {
    return plus (f->expression (x), g->expression (x)); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) + g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) + g->move (z); }
};

TMPL Analytic
operator + (const Analytic& f, const Analytic& g) {
  return Analytic (new sum_analytic_rep<V> (f, g));
}

TMPL_DEF
class diff_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  diff_analytic_rep (const Analytic& f2, const Analytic& g2):
    analytic_rep<V> (f2->s - g2->s), f(f2), g(g2) {}
  generic expression (const generic& x) const {
    return minus (f->expression (x), g->expression (x)); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) + g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) - g->move (z); }
};

TMPL Analytic
operator - (const Analytic& f) {
  return Analytic (new diff_analytic_rep<V> (0, f));
}

TMPL Analytic
operator - (const Analytic& f, const Analytic& g) {
  return Analytic (new diff_analytic_rep<V> (f, g));
}

/******************************************************************************
* Scalar multiplication and division
******************************************************************************/

TMPL_DEF
class scprod_analytic_rep: public Analytic_rep {
  Complex  c;
  Analytic f;
public:
  scprod_analytic_rep (const Complex& c2, const Analytic& f2):
    analytic_rep<V> (c2 * f2->s), c (c2), f (f2) {}
  generic expression (const generic& x) const {
    return times (as_generic (c), f->expression (x)); }
  Positive radius (nat prec) const {
    return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return V::_abs (c) * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return c * f->move (z); }
};

TMPL Analytic
operator * (const Complex& c, const Analytic& f) {
  return Analytic (new scprod_analytic_rep<V> (c, f));
}

TMPL Analytic
operator * (const Analytic& f, const Complex& c) {
  return Analytic (new scprod_analytic_rep<V> (c, f));
}

TMPL Analytic
operator / (const Analytic& f, const Complex& c) {
  if (c == Complex (0)) return 0;
  else return Analytic (new scprod_analytic_rep<V> (1/c, f));
}

TMPL Analytic
operator % (const Analytic& f, const Complex& c) {
  if (c == Complex (0)) return f;
  else return 0;
}

/******************************************************************************
* Multiplication
******************************************************************************/

TMPL_DEF
class prod_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  prod_analytic_rep (const Analytic& f2, const Analytic& g2):
    analytic_rep<V> (f2->s * g2->s), f (f2), g (g2) {}
  generic expression (const generic& x) const {
    return times (f->expression (x), g->expression (x)); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->radius (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) * g->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) * g->move (z); }
};

TMPL Analytic
operator * (const Analytic& f, const Analytic& g) {
  return Analytic (new prod_analytic_rep<V> (f, g));
}

/******************************************************************************
* Division
******************************************************************************/

TMPL_DEF
class quot_analytic_rep: public Analytic_rep {
  Analytic f, g;
public:
  quot_analytic_rep (const Analytic& f2, const Analytic& g2):
    analytic_rep<V> (f2->s / g2->s), f (f2), g (g2) {}
  generic expression (const generic& x) const {
    return over (f->expression (x), g->expression (x)); }
  Positive radius (nat prec) const {
    return min (f->radius (prec), g->root_free_disk (prec)); }
  Positive bound (const Positive& r, nat prec) const {
    return f->bound (r, prec) / g->lower_bound (r, prec); }
  Analytic move (const Complex& z) const {
    return f->move (z) / g->move (z); }
};

TMPL Analytic
operator / (const Analytic& f, const Analytic& g) {
  return Analytic (new quot_analytic_rep<V> (f, g));
}

/******************************************************************************
* Differentiation
******************************************************************************/

TMPL_DEF
class derive_analytic_rep: public Analytic_rep {
  Analytic f;
public:
  derive_analytic_rep (const Analytic& f2):
    analytic_rep<V> (derive (f2->s)), f (f2) {}
  generic expression (const generic& x) const {
    return as_generic ("derive", f->expression (var (f)), var (f)) (x); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    Positive R= radius_between (r, radius (prec));
    return (R / square (R - r)) * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return derive (f->move (z)); }
};

TMPL Analytic
derive (const Analytic& f) {
  return Analytic (new derive_analytic_rep<V> (f));
}

/******************************************************************************
* Integration
******************************************************************************/

TMPL_DEF
class integrate_analytic_rep: public Analytic_rep {
  Analytic f;
  Complex  c;
public:
  integrate_analytic_rep (const Analytic& f2, const Complex& c2):
    analytic_rep<V> (integrate (f2->s) + Series (c2)), f (f2), c (c2) {}
  generic expression (const generic& x) const {
    return as_generic ("integrate", f->expression (var (f)), var (f)) (x); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return r * f->bound (r, prec); }
  Analytic move (const Complex& z) const {
    return integrate (f->move (z), evaluate (z)); }
  Complex eval (const Complex& z) const {
    return evaluate (z); }
};

TMPL Analytic
integrate (const Analytic& f, const Complex& c= 0) {
  return Analytic (new integrate_analytic_rep<V> (f, c));
}

/******************************************************************************
* Exponentiation
******************************************************************************/

TMPL_DEF
class exp_analytic_rep: public Analytic_rep {
  Analytic f;
public:
  exp_analytic_rep (const Analytic& f2):
    analytic_rep<V> (exp (f2->s)), f (f2) {}
  generic expression (const generic& x) const {
    return as_generic ("exp", f->expression (x)); }
  Positive radius (nat prec) const { return f->radius (prec); }
  Positive bound (const Positive& r, nat prec) const {
    return exp (f->bound (r, prec)); }
  Analytic move (const Complex& z) const {
    return exp (f->move (z)); }
};

TMPL Analytic
exp (const Analytic& f) {
  return Analytic (new exp_analytic_rep<V> (f));
}

#undef TMPL_DEF
#undef TMPL
#undef Analytic
#undef Analytic_rep
#undef Positive
#undef Complex
#undef Series
} // namespace mml
#endif // __MML_ANALYTIC_HPP
