
/******************************************************************************
* MODULE     : analytic_eval.hpp
* DESCRIPTION: Evaluation of an analytic functions in a point
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "analytic.hpp"
namespace mml {

/******************************************************************************
* Evaluation in double precision case
******************************************************************************/

template<> double_analytic::_complex
analytic_rep<double_analytic>::evaluate
  (const double_analytic::_complex& z) const
{
  double_analytic::_positive r= double_analytic::_abs (z);
  double_analytic::_positive R= radius_between (r, radius ());
  //cout << "[" << r << ", " << R << "]";
  double_analytic::_positive M= bound (R);
  double_analytic::_positive a= r / R;
  double_analytic::_positive B= M / (1.0 - a);
  double_analytic::_positive C= 2.625223256644e-33; // 2^(-52)
  double_analytic::_complex sum (0.0), pow (1.0);
  for (nat i=0; i<1000; i++) {
    sum += s[i] * pow;
    pow *= z;
    B   *= a;
    if (B < C * abs(sum)) break;
  }
  return sum;
}

/******************************************************************************
* Evaluation in certified precision case
******************************************************************************/

class evaluate_analytic_rep: public var_complex_rep {
private:
  analytic<exact_analytic> f;
  complex z;
  interval M, a;
  
public:
  inline evaluate_analytic_rep (const analytic<exact_analytic>& f2,
				const complex& z2):
    var_complex_rep (weight (z2)),
    f (f2), z (z2) { best= ansatz (); init_magn (); }
  generic expression () const { return f->expression (as_generic (z)); }
  void compute_parameters () {
    interval r = exact_analytic::_abs (z);
    interval R = radius_between (r, f->radius ());
    M = f->bound (R);
    a = r / R;
    //cout << "M= " << M << "\n";
    //cout << "a= " << a << "\n";
  }
  complex_interval ansatz () {
    compute_parameters ();
    complex_interval sum= 0;
    interval prod= 1;
    dyadic asum= 0, rest= upper (M / (interval (1) - a));
    nat n;
    for (n= 0; asum < rest; n++) {
      sum  += mml::sq_approx (f[n], lp_plus_infinity) * prod;
      asum += sq_bound (f[n]) * upper (prod);
      prod *= a;
      rest *= upper (a);
      if (asum == 0 && n > 100) return 0;
    }
    //cout << "[" << n << "]";
    interval delta= segment (-rest, rest);
    return sum + complex_interval (delta, delta);
  }
  void sq_approx (const dyadic& err2) {
    //cout << "approx " << err2 << "\n";
    compute_parameters ();
    dyadic err= err2 >> 1;
    nat n= 1;
    if (!is_infinity (err)) {
      interval n0= log ((interval (1) - a) * interval (err) / M) / log (a);
      n= as_nat (ceil (upper (n0)));
      if (n == 0) n= 1;
    }
    //cout << "n= " << n << "\n";
    complex sum (f[n-1]);
    for (int i= ((int) n)-2; i>=0; i--)
      sum= f[(nat) i] + z * sum;
    //cout << "s= " << sum << "\n";
    interval re= approx_mult (Re (sum), err / dyadic (1.414213563));
    interval im= approx_mult (Im (sum), err / dyadic (1.414213563));
    dyadic   delta_d= bound (M * pow (a, n) / (interval (1) - a));
    interval delta_i= segment (-delta_d, delta_d);
    best= complex_interval (re + delta_i, im + delta_i);
    //cout << "best= " << best << "\n";
  }
};

complex
evaluate_analytic (const analytic<exact_analytic>& f, const complex& z) {
  var_complex u= var_complex (new evaluate_analytic_rep (f, z));
  return (complex) var_complex (new evaluate_analytic_rep (f, z));  
}

template<> exact_analytic::_complex
analytic_rep<exact_analytic>::evaluate
  (const exact_analytic::_complex& z) const
{
  analytic_rep<exact_analytic>* rep=
    (analytic_rep<exact_analytic>*) ((void*) this);
  return evaluate_analytic (analytic<exact_analytic> (rep), z);
}

} // namespace mml
