
/******************************************************************************
* MODULE     : defaults.hpp
* DESCRIPTION: Templates and default implementations of several operations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DEFAULTS_HPP
#define __MML_DEFAULTS_HPP
namespace mml {

/******************************************************************************
* Hashing
******************************************************************************/

inline nat hash (int c) { return (nat) c; }
inline nat hash (nat c) { return (nat) c; }
inline nat hash (char c) { return (nat) c; }
inline nat hash (unsigned char c) { return (nat) c; }

/******************************************************************************
* Arithmetic
******************************************************************************/

#define ARITH_INT_SUGAR(TMPL,T,C)				        \
TMPL inline T operator + (const T& x, int y) { return x + T(y); }	\
TMPL inline T operator + (int x, const T& y) { return T(x) + y; }	\
TMPL inline T operator - (const T& x, int y) { return x - T(y); }	\
TMPL inline T operator - (int x, const T& y) { return T(x) - y; }	\
TMPL inline T operator * (const T& x, int y) { return x * C(y); }	\
TMPL inline T operator * (int x, const T& y) { return C(x) * y; }	\
TMPL inline T operator / (const T& x, int y) { return x / C(y); }	\
TMPL inline T operator / (int x, const T& y) { return T(x) / y; }	\
TMPL inline T operator % (const T& x, int y) { return x % C(y); }	\
TMPL inline T operator % (int x, const T& y) { return T(x) % y; }

#define SHIFT_INT_SUGAR(TMPL,T,C)				        \
TMPL inline T& operator <<= (T& x, int y) { return x= x << y; }	        \
TMPL inline T& operator >>= (T& x, int y) { return x= x >> y; }

#define ARITH_SCALAR_SUGAR(TMPL,T,C)				        \
TMPL inline T operator - (const T& x) { return T(0) - x; }		\
TMPL inline T operator * (const T& x, const C& y) { return x * T(y); }	\
TMPL inline T operator * (const C& x, const T& y) { return T(x) * y; }	\
TMPL inline T operator / (const T& x, const C& y) { return x / T(y); }	\
TMPL inline T operator % (const T& x, const C& y) { return x % T(y); }

template<typename C, typename D> inline C&
operator += (C& x, const D& y) { return x= x+y; }
template<typename C, typename D> inline C&
operator -= (C& x, const D& y) { return x= x-y; }
template<typename C, typename D> inline C&
operator *= (C& x, const D& y) { return x= x*y; }
template<typename C, typename D> inline C&
operator /= (C& x, const D& y) { return x= x/y; }
template<typename C, typename D> inline C&
operator %= (C& x, const D& y) { return x= x%y; }

/******************************************************************************
* Comparison
******************************************************************************/

template<typename C> int
sign (const C& x) {
  if (x>0) return 1;
  if (x<0) return -1;
  return 0;
}

template<typename C> int
compare (const C& x, const C& y) {
  return sign (x-y);
}

template<typename C> C abs (const C& x) { return x>=0? x: -x; }
template<typename C> C min (const C& x, const C& y) { return x<y? x: y; }
template<typename C> C max (const C& x, const C& y) { return x>y? x: y; }

#define COMPARE_SUGAR(TMPL,T)				\
TMPL inline bool operator == (const T& x, const T& y) {	\
  return compare (x,y) == 0; }				\
TMPL inline bool operator != (const T& x, const T& y) {	\
  return compare (x,y) != 0; }				\
TMPL inline bool operator <  (const T& x, const T& y) {	\
  return compare (x,y) <  0; }				\
TMPL inline bool operator <= (const T& x, const T& y) {	\
  return compare (x,y) <= 0; }				\
TMPL inline bool operator >  (const T& x, const T& y) {	\
  return compare (x,y) >  0; }				\
TMPL inline bool operator >= (const T& x, const T& y) {	\
  return compare (x,y) >= 0; }

#define STRICT_COMPARE_SUGAR(TMPL,T)			\
TMPL inline bool operator != (const T& x, const T& y) {	\
  return !(x==y); }					\
TMPL inline bool operator <  (const T& x, const T& y) {	\
  return x<=y && !(x==y); }				\
TMPL inline bool operator >  (const T& x, const T& y) {	\
  return x>=y && !(x==y); }

#define EQUAL_INT_SUGAR(TMPL,T, C)					\
TMPL inline bool operator == (const T& x, int y) { return x == C(y); }	\
TMPL inline bool operator != (const T& x, int y) { return x != C(y); }

#define COMPARE_INT_SUGAR(TMPL,T, C)					\
TMPL inline bool operator == (const T& x, int y) { return x == C(y); }	\
TMPL inline bool operator != (const T& x, int y) { return x != C(y); }	\
TMPL inline bool operator <  (const T& x, int y) { return x <  C(y); }	\
TMPL inline bool operator <= (const T& x, int y) { return x <= C(y); }	\
TMPL inline bool operator >  (const T& x, int y) { return x >  C(y); }	\
TMPL inline bool operator >= (const T& x, int y) { return x >= C(y); }

/******************************************************************************
* Special functions
******************************************************************************/

template<typename C> C
binpow (const C& i, nat n) {
  // FIXME: how to merge with pow?
  if (n <= 1) return n==0? C(1): i;
  C j= binpow (i, n>>1);
  if ((n&1) == 0) return j*j;
  else return j*j*i;
}

template<typename C> C
primitive_root (nat n) {
  assert (false); // not implemented
  return 1;
}

template<typename C> inline C square (const C& x) {
  return x*x; }
template<typename C> inline C sqrt (const C& x) {
  assert (x == C(1)); return 1; }
template<typename C> inline C hypot (const C& x, const C& y) {
  return sqrt (square (x) + square (y)); }
template<typename C, typename D> inline C pow (const C& x, const D& y) {
  assert (x == C(1)); return 1; }
template<typename C> inline C pow (const C& x, const nat& n) {
  return binpow (x, n); }

template<typename C> inline C exp (const C& x) {
  assert (x == C(0)); return 1; }
template<typename C> inline C log (const C& x) {
  assert (x == C(1)); return 0; }
template<typename C> inline C cos (const C& x) {
  assert (x == C(0)); return 1; }
template<typename C> inline C sin (const C& x) {
  assert (x == C(0)); return 0; }
template<typename C> inline C tan (const C& x) {
  return sin (x) / cos (x); }
template<typename C> inline C sec (const C& x) {
  return C(1) / cos (x); }
template<typename C> inline C csc (const C& x) {
  return C(1) / sin (x); }
template<typename C> inline C cot (const C& x) {
  return cos (x) / sin (x); }
template<typename C> inline C atan2 (const C& y, const C& x) {
  if (x > C(0)) return atan (y/x);
  else if (y == C(0)) return x == C(0)? C(0): C(-4) * atan (C(1));
  else return sign (y) * (C(4) * atan (C(1)) - atan (abs (y/x))); }
template<typename C> inline C cosh (const C& x) {
  return (exp (x) + exp (-x)) / C (2); }
template<typename C> inline C sinh (const C& x) {
  return (exp (x) - exp (-x)) / C (2); }
template<typename C> inline C tanh (const C& x) {
  return (exp (x) - exp (-x)) / (exp (x) + exp (-x)); }
template<typename C> inline C sech (const C& x) {
  return C(2) / (exp (x) + exp (-x)); }
template<typename C> inline C csch (const C& x) {
  return C(2) / (exp (x) - exp (-x)); }
template<typename C> inline C coth (const C& x) {
  return (exp (x) + exp (-x)) / (exp (x) - exp (-x)); }
template<typename C> inline C acosh (const C& x) {
  return log (x + sqrt (square (x) - C (1))); }
template<typename C> inline C asinh (const C& x) {
  return log (x + sqrt (square (x) + C (1))); }
template<typename C> inline C atanh (const C& x) {
  return log ((x + C (1)) / (x - C (1))) / C (2); }

/******************************************************************************
* Special instanciations for doubles
******************************************************************************/

template<> inline double sqrt (const double& x) { return ::sqrt (x); }
template<> inline double hypot (const double& x, const double& y) {
  return ::hypot (x, y); }
template<> inline double pow (const double& x, const double& y) {
  return ::pow (x, y); }

template<> inline double exp (const double& x) { return ::exp (x); }
template<> inline double log (const double& x) { return ::log (x); }
template<> inline double cos (const double& x) { return ::cos (x); }
template<> inline double sin (const double& x) { return ::sin (x); }
template<> inline double tan (const double& x) { return ::tan (x); }
template<> inline double cosh (const double& x) { return ::cosh (x); }
template<> inline double sinh (const double& x) { return ::sinh (x); }
template<> inline double tanh (const double& x) { return ::tanh (x); }
template<> inline double atan2 (const double& y, const double& x) {
  return ::atan2 (y, x); }

} // namespace mml
#endif // __MML_DEFAULTS_HPP
