
/******************************************************************************
* MODULE     : complexify.hpp
* DESCRIPTION: Complexified numbers
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_COMPLEXIFY_HPP
#define __MML_COMPLEXIFY_HPP
#include "string.hpp"
namespace mml {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
#define Complexify complexify<C>
TMPL class complexify;
TMPL inline C Re (const Complexify& z);
TMPL inline C Im (const Complexify& z);

/******************************************************************************
* The complexify type
******************************************************************************/

TMPL_DEF
class complexify {
private:
  C re;
  C im;
public:
  inline complexify (): re (0), im (0) {}
  inline complexify (int i): re (i), im (0) {}
  inline complexify (const C& re2, const C& im2= 0): re (re2), im (im2) {}
  friend C Re LESSGTR (const Complexify& z);
  friend C Im LESSGTR (const Complexify& z);

  inline Complexify& operator += (const Complexify& z);
  inline Complexify& operator -= (const Complexify& z);
  inline Complexify& operator *= (const C& x);
  inline Complexify& operator /= (const C& x);
  inline Complexify& operator %= (const C& x);
  inline Complexify& operator *= (const Complexify& z);
  inline Complexify& operator /= (const Complexify& z);
  inline Complexify& operator %= (const Complexify& z);
};

/******************************************************************************
* Basic routines
******************************************************************************/

TMPL inline C Re (const Complexify& z) { return z.re; }
TMPL inline C Im (const Complexify& z) { return z.im; }

TMPL inline nat
hash (const Complexify& z) {
  nat h= hash (Re (z));
  return (h<<1) ^ (h<<3) ^ (h<<9) ^ (h>>29) ^ hash (Im (z));
}

TMPL generic
as_generic (const Complexify& z) {
  return plus (as_generic (Re (z)), times (as_generic (Im (z)), "mathi"));
}

PRINT_SUGAR(TMPL,Complexify)

TMPL inline bool
operator == (const Complexify& z1, const Complexify& z2) {
  return (Re (z1) == Re (z2)) && (Im (z1) == Im (z2));
}

TMPL inline bool
operator != (const Complexify& z1, const Complexify& z2) {
  return (Re (z1) != Re (z2)) || (Im (z1) != Im (z2));
}

/******************************************************************************
* Polar coordinates and complex conjugation
******************************************************************************/

TMPL inline Complexify
polar (const C& r, const C& t) {
  return Complexify (r * cos (t), r * sin (t));
}

TMPL inline C
abs (const Complexify& z) {
  return hypot (Re (z), Im (z));
}

TMPL inline C
arg (const Complexify& z) {
  return atan2 (Im (z), Re (z));
}

TMPL inline C
norm (const Complexify& z) {
  return square (Re (z)) + square (Im (z));
}

TMPL inline Complexify
conj (const Complexify& z) {
  return Complexify (Re (z), -Im (z));
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

TMPL inline Complexify
operator + (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) + Re (z2), Im (z1) + Im (z2));
}

TMPL inline Complexify
operator - (const Complexify& z) {
  return Complexify (-Re (z), -Im (z));
}

TMPL inline Complexify
operator - (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) - Re (z2), Im (z1) - Im (z2));
}

TMPL inline Complexify
operator * (const C& x, const Complexify& z) {
  return Complexify (x * Re (z), x * Im (z));
}

TMPL inline Complexify
operator * (const Complexify& z, const C& x) {
  return Complexify (Re (z) * x, Im (z) * x);
}

TMPL inline Complexify
operator / (const Complexify& z, const C& x) {
  return Complexify (Re (z) / x, Im (z) / x);
}

TMPL inline Complexify
operator / (const C& x, const Complexify& z) {
  C a= x / norm (z);
  return Complexify (a * Re (z), -a * Im (z));
}

TMPL inline Complexify
operator % (const Complexify& z, const C& x) {
  return Complexify (Re (z) % x, Im (z) % x);
}

TMPL inline Complexify
operator % (const C& x, const Complexify& z) {
  if (z == Complexify (0)) return x;
  else return 0;
}

TMPL inline Complexify
operator * (const Complexify& z1, const Complexify& z2) {
  return Complexify (Re (z1) * Re (z2) - Im (z1) * Im (z2),
		     Re (z1) * Im (z2) + Im (z1) * Re (z2));
}

TMPL inline Complexify
operator / (const Complexify& z1, const Complexify& z2) {
  C a= norm (z2);
  return Complexify ((Re (z1) * Re (z2) + Im (z1) * Im (z2)) / a,
		     (Im (z1) * Re (z2) - Re (z1) * Im (z2)) / a);
}

TMPL inline Complexify
operator % (const Complexify& z1, const Complexify& z2) {
  if (z2 == Complexify (0)) return z1;
  else return 0;
}

/******************************************************************************
* In place arithmetic operations
******************************************************************************/

TMPL inline Complexify&
Complexify::operator += (const Complexify& z) {
  re += z.re;
  im += z.im;
  return *this;
}

TMPL inline Complexify&
Complexify::operator -= (const Complexify& z) {
  re -= z.re;
  im -= z.im;
  return *this;
}

TMPL inline Complexify&
Complexify::operator *= (const C& x) {
  re *= x;
  im *= x;
  return *this;
}

TMPL inline Complexify&
Complexify::operator /= (const C& x) {
  re /= x;
  im /= x;
  return *this;
}

TMPL inline Complexify&
Complexify::operator %= (const C& x) {
  re %= x;
  im %= x;
  return *this;
}

TMPL inline Complexify&
Complexify::operator *= (const Complexify& z) {
  C RE= re;
  re= re * z.re - im * z.im;
  im= RE * z.im + im * z.re;
  return *this;
}

TMPL inline Complexify&
Complexify::operator /= (const Complexify& z) {
  C RE= re;
  C a = norm (z);
  re= (re * z.re + im * z.im) / a;
  im= (im * z.re - RE * z.im) / a;
  return *this;
}

TMPL inline Complexify&
Complexify::operator %= (const Complexify& z) {
  if (z != Complexify (0)) re= im= 0;
  return *this;
}

/******************************************************************************
* Special treatment of % for doubles
******************************************************************************/

inline complexify<double>
operator % (const complexify<double>& z, const double& x) {
  if (x == 0.0) return z;
  else return complexify<double> (0.0, 0.0);
}

template<> inline complexify<double>&
complexify<double>::operator %= (const double& x) {
  if (x != 0.0) re= im= 0.0;
  return *this;
}

/******************************************************************************
* Exponentiation and logarithms
******************************************************************************/

TMPL inline Complexify
exp (const Complexify& z) {
  return polar (exp (Re (z)), Im (z));
}

TMPL inline Complexify
log (const Complexify& z) {
  return Complexify (log (abs (z)), arg (z));
}

TMPL inline Complexify
cos (const Complexify& z) {
  return Complexify (cos (Re (z)) * cosh (Im (z)),
		     -sin (Re (z)) * sinh (Im (z)));
}

TMPL inline Complexify
cosh (const Complexify& z) {
  return Complexify (cosh (Re (z)) * cos (Im (z)),
		     sinh (Re (z)) * sin (Im (z)));
}

TMPL inline Complexify
sin (const Complexify& z) {
  return Complexify (sin (Re (z)) * cosh (Im (z)),
		     cos (Re (z)) * sinh (Im (z)));
}

TMPL inline Complexify
sinh (const Complexify& z) {
  return Complexify (sinh (Re (z)) * cos (Im (z)),
		     cosh (Re (z)) * sin (Im (z)));
}

TMPL inline Complexify
pow (const Complexify& z, const Complexify& y) {
  C r= log (abs (z));
  C t= arg (z);
  return polar (exp (r * Re (y) - Im (y) * t),
		Im (y) * r + Re (y) * t);
}

TMPL inline Complexify
pow (const Complexify& z, const C& y) {
  return exp (y * log (z));
}

TMPL inline Complexify
pow (const C& z, const Complexify& y) {
  return exp (y * log (z));
}

/******************************************************************************
* Syntactic sugar
******************************************************************************/

ARITH_INT_SUGAR(TMPL,Complexify,C);
EQUAL_INT_SUGAR(TMPL,Complexify,Complexify);

#undef TMPL_DEF
#undef TMPL
#undef Complexify
} // namespace mml
#endif // __MML_COMPLEXIFY_HPP
