
/******************************************************************************
* MODULE     : interval.hpp
* DESCRIPTION: Intervals with a single precision radius
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* Intervals are either represented by their borders (in low precision) or
* by their center and radius (in high precision). Here low precision means
* a precision of <= DEFAULT_LOW bits. In high precision, the radius is still
* stored in a low precision number. Moreover, the "step" of the radius is
* always larger or equal to the "step" of the center (the step is the binary
* exponent plus the bit precision). If the step of the radius gets too large,
* then the precision of the center is automatically decreased.
* In the high precision representation, things are also normalized in
* such a way that we always have radius <= |center| / 2^DEFAULT_LOW.
* This can be helpful for making error estimations.
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_INTERVAL_HPP
#define __MML_INTERVAL_HPP
#include "dyadic.hpp"
namespace mml {

typedef int (*mpfr_cst_algo) (mpfr_t, mp_rnd_t);

#define CHECK_ERROR(r) \
  if (mpfr_nan_p ((r)->mid)) return r->finish_error ();
#define CHECK_UNDERFLOW(r) \
  if (mpfr_sgn ((r)->mid) == 0) return r->finish_underflow ();
#define CHECK_SIGNED_UNDERFLOW(r,sgn) \
  if (mpfr_sgn ((r)->mid) == 0) return r->finish_signed_underflow (sgn);
#define CHECK_OVERFLOW(r) \
  if (mpfr_inf_p ((r)->mid)) return r->finish_overflow ();

class interval_rep;
class interval {
INDIRECT_PROTO (interval, interval_rep)
public:
  inline interval (signed int i= 0, mp_prec_t prec= DEFAULT_PREC);
  inline interval (unsigned int i, mp_prec_t prec= DEFAULT_PREC);
  inline interval (signed short int i, mp_prec_t prec= DEFAULT_PREC);
  inline interval (unsigned short int i, mp_prec_t prec= DEFAULT_PREC);
  inline interval (signed long int i, mp_prec_t prec= DEFAULT_PREC);
  inline interval (unsigned long int i, mp_prec_t prec= DEFAULT_PREC);
  inline interval (float x, mp_prec_t prec= DEFAULT_PREC);
  inline interval (double x, mp_prec_t prec= DEFAULT_PREC);
  inline interval (const dyadic& x, mp_prec_t prec= DEFAULT_PREC);
  inline interval (const integer& x, mp_prec_t prec= DEFAULT_PREC);
  inline interval (const rational& x, mp_prec_t prec= DEFAULT_PREC);

  static interval nanin;
  static interval epsilon;
  static interval fuzzy;
  static interval plus_infinity;
  static interval minus_infinity;
  static interval log2 (mp_prec_t prec= DEFAULT_LOW);
  static interval pi (mp_prec_t prec= DEFAULT_LOW);
  static interval euler (mp_prec_t prec= DEFAULT_LOW);
  //static interval catalan (mp_prec_t prec= DEFAULT_LOW);

  friend nat hash (const interval& x);
};

class interval_rep REP_STRUCT {
private:
  union {
    mpfr_t mid;
    mpfr_t min;
  };
  union {
    mpfr_t rad;
    mpfr_t max;
  };

private:
  inline interval_rep () {}
  inline friend interval_rep* raw_interval (mp_prec_t prec= DEFAULT_LOW) {
    interval_rep* rep= new interval_rep ();
    mpfr_init2 (rep->mid, prec);
    mpfr_init2 (rep->rad, DEFAULT_LOW);
    return rep; }
  interval finish_high ();
  interval finish_underflow ();
  interval finish_signed_underflow (int sgn);
  interval finish_overflow ();
  interval finish_error ();

public:
  // Constructors and destructors
  friend class interval;
  interval_rep (signed long int i, mp_prec_t prec);
  interval_rep (unsigned long int i, mp_prec_t prec);
  interval_rep (double x, mp_prec_t prec);
  interval_rep (const dyadic& x, mp_prec_t prec);
  inline ~interval_rep () { mpfr_clear (mid); mpfr_clear (rad); }

  // Frequently used predicates
  inline friend bool is_error (const interval& x) {
    return mpfr_nan_p (x->mid) != 0; }
  inline friend bool is_fuzzy (const interval& x) {
    return (mpfr_inf_p (x->min) != 0) || (mpfr_inf_p (x->max) != 0); }
  inline friend bool is_infinity (const interval& x) {
    return (mpfr_inf_p (x->min) != 0 && mpfr_sgn (x->min) > 0) ||
           (mpfr_inf_p (x->max) != 0 && mpfr_sgn (x->max) < 0); }
  inline friend bool is_epsilon (const interval& x) {
    return (mpfr_sgn (x->max) == 0) && (mpfr_sgn (x->min) == 0); }
  inline friend bool is_exact (const interval& x) {
    if (mpfr_get_prec (x->mid) <= DEFAULT_LOW)
      return mpfr_cmp (x->min, x->max) == 0;
    else return mpfr_sgn (x->rad) == 0; }
  inline friend bool is_non_zero_exact (const interval& x) {
    if (mpfr_get_prec (x->mid) <= DEFAULT_LOW)
      return mpfr_cmp (x->min, x->max) == 0 && mpfr_sgn (x->min) != 0;
    else return mpfr_sgn (x->rad) == 0; }
  inline friend bool is_signed (const interval& x) {
    return radius (x) < (abs (center (x)) >> 2); }
  inline friend bool has_low_prec (const interval& x) {
    return mpfr_get_prec (x->mid) <= DEFAULT_LOW; }
  inline friend bool has_high_prec (const interval& x) {
    return mpfr_get_prec (x->mid) > DEFAULT_LOW; }
  inline friend bool has_huge_prec (const interval& x) {
    return mpfr_get_prec (x->mid) > (DEFAULT_LOW << 1); }

  // Changing precision
  friend interval reduce (const interval& x);
  friend interval reduce (const interval& x, mp_prec_t prec);
  friend interval extend (const interval& x, mp_prec_t prec);
  friend interval truncate (const interval& x, mp_prec_t prec);

  // Interface with dyadic numbers
  friend interval segment (const dyadic& min, const dyadic& max);
  friend interval ball (const dyadic& mid, const dyadic& rad);
  friend dyadic center (const interval& x);
  friend dyadic radius (const interval& x);
  friend dyadic lower (const interval& x);
  friend dyadic upper (const interval& x);
  friend dyadic bound (const interval& x);
  friend bool wider (const interval& x, const dyadic& rad);
  inline friend dyadic magnitude (const interval& x) {
    return abs (center (x)); }

  // Low level access functions
  inline friend mp_prec_t prec (const interval& x) {
    return (mp_exp_t) mpfr_get_prec (x->mid); }
  inline friend mp_exp_t expo_mid (const interval& x) {
    return mpfr_get_exp (x->mid); }
  inline friend mp_exp_t step_mid (const interval& x) {
    return mpfr_get_exp (x->mid) - mpfr_get_prec (x->mid); }
  inline friend mp_exp_t expo (const interval& x) {
    if (mpfr_get_prec (x->mid) > DEFAULT_LOW)
      return mpfr_get_exp (x->mid) + 1;
    else if ((mpfr_sgn (x->min) == 0) && (mpfr_sgn (x->max) == 0))
      return mpfr_get_emin ();
    else return mml::max (mpfr_get_exp (x->min), mpfr_get_exp (x->max)); }
  inline friend mp_exp_t step (const interval& x) {
    if (mpfr_get_prec (x->mid) > DEFAULT_LOW)
      return mpfr_get_exp (x->mid) + 1 - mpfr_get_prec (x->mid);
    else if ((mpfr_sgn (x->min) == 0) && (mpfr_sgn (x->max) == 0))
      return mpfr_get_emin ();
    else return
      mml::max (mpfr_get_exp (x->min), mpfr_get_exp (x->max)) - DEFAULT_LOW; }

  // Input/output
  friend ostream& operator << (ostream& out, const interval& x);
  friend generic  as_generic (const interval& x);

  // Basic arithmetic
  friend interval operator - (const interval& x);
  friend interval operator + (const interval& x, const interval& y);
  friend interval operator - (const interval& x, const interval& y);
  friend interval operator * (const interval& x, const interval& y);
  friend interval operator / (const interval& x, const interval& y);
  friend interval operator << (const interval& x, mp_exp_t shift);
  friend interval operator >> (const interval& x, mp_exp_t shift);
  friend interval abs (const interval& x);
  friend interval sqrt (const interval& x);

  // Special functions
  friend interval exp (const interval& x);
  friend interval log (const interval& x);
  friend interval cos (const interval& x);
  friend interval sin (const interval& x);
  friend interval acos (const interval& x);
  friend interval asin (const interval& x);
  friend interval atan (const interval& x);
  friend interval atan2 (const interval& y, const interval& x);

  // miscellaneous functions
  friend interval min (const interval& x, const interval& y);
  friend interval max (const interval& x, const interval& y);
  friend interval constant_interval (mpfr_cst_algo r, mp_prec_t prec);
  inline friend nat hash (const interval& x) {
    __mpfr_struct &rep= x.rep->mid[0];
    if (rep._mpfr_prec == 0) return 0;
    return ((nat) (rep._mpfr_d[0])) ^ ((nat) (rep._mpfr_exp << 2));
  }
};

INDIRECT_IMPL (interval, interval_rep)
inline interval::interval (signed int i, mp_prec_t prec):
  rep (new interval_rep ((signed long int) i, prec)) {}
inline interval::interval (unsigned int i, mp_prec_t prec):
  rep (new interval_rep ((unsigned long int) i, prec)) {}
inline interval::interval (signed short int i, mp_prec_t prec):
  rep (new interval_rep ((signed long int) i, prec)) {}
inline interval::interval (unsigned short int i, mp_prec_t prec):
  rep (new interval_rep ((unsigned long int) i, prec)) {}
inline interval::interval (signed long int i, mp_prec_t prec):
  rep (new interval_rep (i, prec)) {}
inline interval::interval (unsigned long int i, mp_prec_t prec):
  rep (new interval_rep (i, prec)) {}
inline interval::interval (float x, mp_prec_t prec):
  rep (new interval_rep ((double) x, prec)) {}
inline interval::interval (double x, mp_prec_t prec):
  rep (new interval_rep (x, prec)) {}
inline interval::interval (const dyadic& x, mp_prec_t prec):
  rep (new interval_rep (x, prec)) {}
inline interval::interval (const integer& i, mp_prec_t prec):
  rep (new interval_rep (dyadic (i, prec), prec)) {}
inline interval::interval (const rational& x, mp_prec_t prec):
  rep (new interval_rep (dyadic (x, prec), prec)) {}

inline interval as_interval (const dyadic& x) {
  return interval (x, prec (x)); }
inline interval as_interval (const integer& i) {
  return as_interval (dyadic (i, (1+limb_size (i)) * BITS_PER_LIMB)); }

inline bool operator < (const interval& x, const interval& y) {
  return upper (x) < lower (y); }
inline bool operator > (const interval& x, const interval& y) {
  return lower (x) > upper (y); }
inline bool operator != (const interval& x, const interval& y) {
  return (upper (x) < lower (y)) || (lower (x) > upper (y)); }
inline bool operator <= (const interval& x, const interval& y) {
  return lower (x) <= upper (y); }
inline bool operator >= (const interval& x, const interval& y) {
  return upper (x) >= lower (y); }
inline bool operator == (const interval& x, const interval& y) {
  return (lower (x) <= upper (y)) && (upper (x) >= lower (y)); }

}
#endif // __MML_INTERVAL_HPP
