
/******************************************************************************
* MODULE     : interval.cpp
* DESCRIPTION: Intervals with a single precision radius
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "interval_impl.hpp"
namespace mml {

interval interval::nanin=
  segment (dyadic::nan (DEFAULT_LOW),
	   dyadic::nan (DEFAULT_LOW));
interval interval::epsilon=
  segment (-dyadic (0, DEFAULT_LOW),
	   dyadic (0, DEFAULT_LOW));
interval interval::fuzzy=
  segment (dyadic::minus_infinity (DEFAULT_LOW),
	   dyadic::plus_infinity (DEFAULT_LOW));
interval interval::plus_infinity=
  segment (dyadic::plus_infinity (DEFAULT_LOW),
	   dyadic::plus_infinity (DEFAULT_LOW));
interval interval::minus_infinity=
  segment (dyadic::minus_infinity (DEFAULT_LOW),
	   dyadic::minus_infinity (DEFAULT_LOW));

/******************************************************************************
* Finish the computation of an interval
******************************************************************************/

interval
interval_rep::finish_high () {
  // This routine should be called just after the raw computation
  // of a high precision interval. It checks whether any exceptions
  // were raised and whether any precision loss occured.
  // In case of precision loss, the precision is decreased and
  // we possibly switch to the low precision representation.
  // Precision loss may both be due to rounding errors and cancellations.

  // Zero center or infinite radius
  if ((mpfr_sgn (mid) == 0) || (mpfr_inf_p (rad) != 0)) {
    mpfr_set_prec (mid, DEFAULT_LOW);
    mpfr_neg (min, rad, GMP_RNDD);
    return this;
  }

  // Zero radius
  if (mpfr_sgn (rad) == 0)
    return this;

  // Precision becomes low
  mp_exp_t expo_mid= mpfr_get_exp (mid);
  mp_exp_t expo_rad= mpfr_get_exp (rad);
  if (expo_mid <= expo_rad + (mp_exp_t) (DEFAULT_LOW+1)) {
    mpfr_prec_round (mid, DEFAULT_LOW, GMP_RNDN);
    mpfr_nextabove (rad);
    direct_dyadic aux;
    mpfr_add (*aux, mid, rad, GMP_RNDU);
    mpfr_sub (min, mid, rad, GMP_RNDD);
    mpfr_set (max, *aux, GMP_RNDN);
    return this;
  }

  // Some precision loss, but the precision remains high
  mp_exp_t step_mid= expo_mid - mpfr_get_prec (mid);
  mp_exp_t step_rad= expo_rad - mpfr_get_prec (rad);
  if (step_mid < step_rad) {
    mpfr_prec_round (mid, expo_mid - expo_rad + DEFAULT_LOW, GMP_RNDN);
    mpfr_nextabove (rad);
  }

  return this;
}

interval
interval_rep::finish_error () {
  mpfr_clear (mid);
  mpfr_init2 (min, DEFAULT_LOW);
  mpfr_clear_flags ();
  return this;
}

interval
interval_rep::finish_underflow () {
  mpfr_clear (mid);
  mpfr_init2 (min, DEFAULT_LOW);
  mpfr_set_si (max, 0, GMP_RNDU);
  mpfr_neg (min, max, GMP_RNDD);
  mpfr_nextbelow (min);
  mpfr_nextabove (max);
  mpfr_clear_flags ();
  return this;
}

interval
interval_rep::finish_signed_underflow (int sgn) {
  mpfr_clear (mid);
  mpfr_init2 (min, DEFAULT_LOW);
  if (sgn > 0) {
    mpfr_set_si (min, 0, GMP_RNDU);
    mpfr_set_si (max, 0, GMP_RNDU);
    mpfr_nextabove (max);
  }
  else {
    mpfr_set_si (min, 0, GMP_RNDD);
    mpfr_set_si (max, 0, GMP_RNDD);
    mpfr_nextbelow (min);
  }
  mpfr_clear_flags ();
  return this;
}

interval
interval_rep::finish_overflow () {
  int sgn= sign (mid);
  mpfr_clear (mid);
  mpfr_init2 (min, DEFAULT_LOW);
  mpfr_set_inf (min, sgn);
  mpfr_set_inf (max, sgn);
  if (sgn > 0) mpfr_nextbelow (min);
  else mpfr_nextabove (max);
  mpfr_clear_flags ();
  return this;
}

} // namespace mml
