
/******************************************************************************
* MODULE     : interval_test.cc
* DESCRIPTION: Test interval numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "interval.hpp"
using namespace mml;

void
test_special () {
  cout << "*** Special values\n\n";
  cout << "epsilon\t= " << interval::epsilon << "\n";
  cout << "fuzzy\t= " << interval::fuzzy << "\n";
  cout << "\n";
}

void
test_add_simple (int prec) {
  cout << "*** Additive arithmetic at precision " << prec << "\n\n";
  interval x= interval (2, prec);
  interval y= interval (7, prec);
  cout << "x\t= " << x << "\n";
  cout << "y\t= " << y << "\n";
  cout << "-x\t= " << (-x) << "\n";
  cout << "x+x\t= " << (x+x) << "\n";
  cout << "x+y\t= " << (x+y) << "\n";
  cout << "x-x\t= " << (x-x) << "\n";
  cout << "x-y\t= " << (x-y) << "\n";
  cout << "(x-x)+y\t= " << ((x-x)+y) << "\n";
  cout << "\n";
}

void
test_mul_simple (int prec) {
  cout << "*** Multiplicative arithmeric at precision " << prec << "\n\n";
  interval x= interval (2, prec) + interval (1, prec);
  cout << "x\t= " << x << "\n";
  cout << "x * x\t= " << x*x << "\n";
  cout << "x * -x\t= " << x*(-x) << "\n";
  cout << "-x * x\t= " << (-x)*x << "\n";
  cout << "-x * -x\t= " << (-x)*(-x) << "\n";
  cout << "x / x\t= " << (x/x) << "\n";
  cout << "x / -x\t= " << (x/(-x)) << "\n";
  cout << "-x / x\t= " << ((-x)/x) << "\n";
  cout << "-x / -x\t= " << ((-x)/(-x)) << "\n";
  cout << "x*x / x\t= " << ((x*x)/x) << "\n";
  cout << "\n";
}

void
test_other_ops (int prec) {
  cout << "*** Other operations " << prec << "\n\n";
  interval x= interval (2, prec);
  interval y= segment (-3, 2);
  cout << "x\t= " << x << "\n";
  cout << "y\t= " << y << "\n";
  cout << "x << 5\t= " << (x << 5) << "\n";
  cout << "x >> 3\t= " << (x >> 3) << "\n";
  cout << "y << 5\t= " << (y << 5) << "\n";
  cout << "abs(x)\t= " << (abs (x)) << "\n";
  cout << "abs(-x)\t= " << (abs (-x)) << "\n";
  cout << "abs(y)\t= " << (abs (y)) << "\n";
  cout << "\n";
}

void
test_mixed_prec (int prec_lo, int prec_hi) {
  cout << "*** Mixed precisions " << prec_lo << ", " << prec_hi << "\n\n";
  interval x= interval (2, prec_lo);
  interval y= interval (7, prec_hi);
  interval p= interval (2.0e-100, prec_lo);
  interval q= interval (7.0e-100, prec_hi);
  cout << "x\t= " << x << "\n";
  cout << "y\t= " << y << "\n";
  cout << "x+y\t= " << (x+y) << "\n";
  cout << "x-y\t= " << (x-y) << "\n";
  cout << "x*y\t= " << (x*y) << "\n";
  cout << "x/y\t= " << (x/y) << "\n";
  cout << "p\t= " << p << "\n";
  cout << "q\t= " << q << "\n";
  cout << "p+y\t= " << (p+y) << "\n";
  cout << "p-y\t= " << (p-y) << "\n";
  cout << "p*y\t= " << (p*y) << "\n";
  cout << "p/y\t= " << (p/y) << "\n";
  cout << "x+q\t= " << (x+q) << "\n";
  cout << "x-q\t= " << (x-q) << "\n";
  cout << "x*q\t= " << (x*q) << "\n";
  cout << "x/q\t= " << (x/q) << "\n";
  cout << "\n";
}

void
test_mul_progression (int prec) {
  cout << "*** Multiplicative progression at precision " << prec << "\n\n";
  int i;
  interval x (3, prec);
  interval y= x;
  for (i=0; i<50; i++) {
    cout << i << "\t" << y << "\n";
    y *= x;
  }
  cout << "\n";
}

void
test_mul_overflow (int prec) {
  cout << "*** Multiplicative overflows at precision " << prec << "\n\n";
  int i;
  interval x (3, prec);
  for (i=0; i<50; i++) {
    cout << i << "\t" << x << "\n";
    x *= x;
  }
  cout << "\n";
}

void
test_mul_underflow (int prec) {
  cout << "*** Multiplicative underflows at precision " << prec << "\n\n";
  int i;
  interval x (0.3, prec);
  for (i=0; i<50; i++) {
    cout << i << "\t" << x << "\n";
    x *= x;
  }
  cout << "\n";
}

void
test_muller (int prec) {
  cout << "*** Muller's sequence at precision " << prec << "\n\n";
  int i;
  interval a (99, prec);
  interval b (-600, prec);
  interval d (-7, prec);
  interval x (4, prec);
  for (i=0; i<50; i++) {
    cout << i << "\t" << x << "\n";
    /*
    interval t1= a*x;
    cout << "t1= " << t1 << "\n";
    interval t2= t1+b;
    cout << "t2= " << t2 << "\n";
    interval t3= x+d;
    cout << "t3= " << t3 << "\n";
    interval t4= t2/t3;
    cout << "t4= " << t4 << "\n";
    x= t4;
    */
    x= (a*x + b) / (x + d);
  }
  cout << "\n";
}

void
test_special_funcs (int prec) {
  cout << "*** Special functions at precision " << prec << "\n\n";
  interval x (1, prec);
  interval y (3e20, prec);
  interval z= x - x/y;
  cout << "log (2)\t\t= " << interval::log2 (prec) << "\n";
  cout << "pi\t\t= " << interval::pi (prec) << "\n";
  cout << "euler\t\t= " << interval::euler (prec) << "\n";
  cout << "x\t\t= " << x << "\n";
  cout << "y\t\t= " << y << "\n";
  cout << "z\t\t= " << z << "\n";
  cout << "sqrt (x)\t= " << sqrt (x) << "\n";
  cout << "sqrt (y)\t= " << sqrt (y) << "\n";
  cout << "exp (x)\t\t= " << exp (x) << "\n";
  cout << "exp (y)\t\t= " << exp (y) << "\n";
  cout << "exp (-x)\t= " << exp (-x) << "\n";
  cout << "exp (-y)\t= " << exp (-y) << "\n";
  cout << "log (x)\t\t= " << log (x) << "\n";
  cout << "log (y)\t\t= " << log (y) << "\n";
  cout << "cos (x)\t\t= " << cos (x) << "\n";
  cout << "cos (y)\t\t= " << cos (y) << "\n";
  cout << "sin (x)\t\t= " << sin (x) << "\n";
  cout << "sin (y)\t\t= " << sin (y) << "\n";
  cout << "tan (x)\t\t= " << tan (x) << "\n";
  cout << "tan (y)\t\t= " << tan (y) << "\n";
  cout << "arc sin (1/2)\t= " << asin (x>>1) << "\n";
  cout << "arc sin (z)\t= " << asin (z) << "\n";
  cout << "arc cos (1/2)\t= " << acos (x>>1) << "\n";
  cout << "arc cos (z)\t= " << acos (z) << "\n";
  cout << "arc tan (x)\t= " << atan (x) << "\n";
  cout << "arc tan (y)\t= " << atan (y) << "\n";
  cout << "arc tan (x)\t= " << atan (x) << "\n";
  cout << "atan2 (x, y)\t= " << atan2 (x, y) << "\n";
  cout << "atan2 (y, x)\t= " << atan2 (y, x) << "\n";
  cout << "atan2 (0, -x)\t= " << atan2 (interval (0), -x) << "\n";
  cout << "min (x, y)\t= " << min (x, y) << "\n";
  cout << "max (x, y)\t= " << max (x, y) << "\n";
  cout << "\n";
}

void
test_all (int prec) {
  test_add_simple (prec);
  test_mul_simple (prec);
  test_other_ops (prec);
  test_mul_progression (prec);
  test_mul_overflow (prec);
  test_mul_underflow (prec);
  test_muller (prec);
  test_special_funcs (prec);
}

int
main () {
  // cout << set_interval_by_bounds (true);

  test_special ();
  test_all (30);
  test_all (100);
  test_all (200);
  test_mixed_prec (30, 100);

  cout << "infty\t\t= " << interval::plus_infinity << "\n";
  cout << "-infty\t\t= " << interval::minus_infinity << "\n";
  cout << "infty+1\t\t= " << interval::plus_infinity + interval (1) << "\n";
  cout << "1-infty\t\t= " << interval (1) - interval::plus_infinity << "\n";
  cout << "1+(-infty)\t= " << interval (1) + interval::minus_infinity << "\n";
  cout << "2 infty\t\t= " << interval (2) * interval::plus_infinity << "\n";
  cout << "1/infty\t\t= " << interval (1) / interval::plus_infinity << "\n";
  cout << "infty^2\t\t= " <<
    interval::plus_infinity * interval::plus_infinity << "\n";
  cout << "(-infty) infty\t= " <<
    interval::minus_infinity * interval::plus_infinity << "\n";
  cout << "(-infty)^2\t= " <<
    interval::minus_infinity * interval::minus_infinity << "\n";
  cout << "infty/infty\t= " <<
    interval::plus_infinity / interval::plus_infinity << "\n";
  cout << "(-infty)/infty\t= " <<
    interval::minus_infinity / interval::plus_infinity << "\n";
  cout << "(-infty)/2\t= " << interval::minus_infinity / interval (2) << "\n";
  cout << "|infty|\t\t= " << abs (interval::plus_infinity) << "\n";
  cout << "|-infty|\t= " << abs (interval::minus_infinity) << "\n";
  cout << "center(infty)\t= " << center (interval::plus_infinity) << "\n";
  cout << "center(-infty)\t= " << center (interval::minus_infinity) << "\n";
  cout << "radius(infty)\t= " << radius (interval::plus_infinity) << "\n";
  cout << "radius(-infty)\t= " << radius (interval::minus_infinity) << "\n";
  cout << "infty << 3\t= " << (interval::plus_infinity << 3) << "\n";
  cout << "infty >> 3\t= " << (interval::plus_infinity >> 3) << "\n";
  cout << "(-infty) << 3\t= " << (interval::minus_infinity << 3) << "\n";
  cout << "(-infty) >> 3\t= " << (interval::minus_infinity >> 3) << "\n";
  return 0;
}
