
/******************************************************************************
* MODULE     : dense_matrix.hpp
* DESCRIPTION: Dense matrices
* COPYRIGHT  : (C) 2004  Olivier Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DENSE_MATRIX_HPP
#define __MML_DENSE_MATRIX_HPP
#include "matrix.hpp"
#include "block.hpp"

namespace mml {
#define TMPL template<typename C, typename M>
#define Matrix matrix<C,dense<M> >
#define Matrix_rep matrix_rep<C,dense<M> >
TMPL inline nat cols (const Matrix& m);
TMPL inline nat rows (const Matrix& m);
TMPL inline C* tab (Matrix& m);
TMPL inline const C* tab (const Matrix& m);

/******************************************************************************
* Dense matrices of arbitrary sizes
******************************************************************************/

template<typename C, typename M>
class Matrix_rep REP_STRUCT {
  C*  a;
  nat nr;
  nat nc;
public:
  inline matrix_rep (C* a2, nat nr2, nat nc2): a (a2), nr (nr2), nc (nc2) {}
  inline ~matrix_rep () { delete[] a; }
  friend class Matrix;
  friend nat cols LESSGTR (const Matrix& m);
  friend nat rows LESSGTR (const Matrix& m);
  friend C* tab LESSGTR (Matrix& m);
  friend const C* tab LESSGTR (const Matrix& m);
};

template<typename C, typename M>
class Matrix {
INDIRECT_PROTO_2 (matrix, matrix_rep, C, dense<M>)
public:
  matrix () {
    nat nc = dense<M>::def_rows;
    nat nr = dense<M>::def_cols;
    C* tab= new C[nr*nc]; 
    rep= new Matrix_rep (tab, nr, nc);
  }

  matrix (int x) {
    nat nr= dense<M>::def_rows;
    nat nc= dense<M>::def_cols;
    C* tab= new C[nr*nc]; 
    rep= new Matrix_rep (tab, nr, nc);
    block_set (rep->a, C(x), nr, nc);
  }

  matrix (const C& x, nat nr=dense<M>::def_rows, nat nc=dense<M>::def_cols) {
    C* tab= new C[nr*nc]; 
    rep= new Matrix_rep (tab, nr, nc);
    block_set (rep->a, x, nr, nc);
  }

  matrix (C* tab, nat nr=dense<M>::def_rows, nat nc=dense<M>::def_cols) {
    rep= new Matrix_rep (tab, nr, nc);
  }

  matrix (const iterator<C>& it) {
    nat i,j, nr= dense<M>::def_rows, nc= dense<M>::def_cols;
    if (nr==0) nr=1;
    C* tab= new C[nr*nc];
    for (i=0; i<nr; i++, ++it)
      for (j=0; j<nc; j++, ++it)
	tab[i*nc+j]= *it;
    rep= new Matrix_rep (tab, nc, nr);
  }
  inline C& ref (nat i, nat j) { return rep->a[i*cols(*this)+j]; }
  inline C operator () (nat i, nat j) const { return rep->a[i*cols(*this)+j]; }
};
INDIRECT_IMPL_2 (matrix, matrix_rep, typename C, C, typename M, dense<M>)

TMPL inline nat rows (const Matrix& m) { return m->nr; }
TMPL inline nat cols (const Matrix& m) {return m->nc; }
TMPL inline C* tab (Matrix& m) { m.secure(); return m->a; }
TMPL inline const C* tab (const Matrix& m) { return m->a; }

/******************************************************************************
* Fixed size dense vectors
******************************************************************************/

#define RMatrix matrix<C,dense<rectangular<naive,RS,CS> > >
#define RMatrix_rep matrix_rep<C,dense<rectangular<naive,RS,CS> > >

template<typename C, nat RS, nat CS>
class RMatrix_rep REP_STRUCT {
  C* a;
public:
  inline matrix_rep (C* a2, nat dummy1, nat dummy2):
    a (a2) { (void) dummy1; (void) dummy2; }
  inline virtual ~matrix_rep () { delete[] a; }
  friend class RMatrix;
  friend nat cols LESSGTR (const RMatrix& m);
  friend nat rows LESSGTR (const RMatrix& m);
  friend C*  tab  LESSGTR (const RMatrix& m);
};

template<typename C, nat RS, nat CS> inline nat
cols (const RMatrix& m) { return CS; }
template<typename C, nat RS, nat CS> inline nat
rows (const RMatrix& m) { return RS; }

#undef RMatrix
#undef RMatrix_rep

/******************************************************************************
* Routines on dense vectors
******************************************************************************/

#define MATRIX_UNARY(name,op)			\
TMPL Matrix					\
operator op (const Matrix& x) {			\
  nat nrows= rows (x);				\
  nat ncols= cols (x);				\
  C* r= new C[nrows*ncols];			\
  name<C> (r, tab (x), nrows*ncols);		\
  return Matrix (r, nrows, ncols);		\
}

#define MATRIX_BINARY(name,op)				\
TMPL Matrix						\
operator op (const Matrix& x, const Matrix& y) {	\
  nat nrows= rows (x);					\
  assert (rows (y) == nrows);				\
  nat ncols= cols(x);					\
  assert (cols(y) == ncols);				\
  C* r= new C[nrows*ncols];				\
  name<C> (r, tab (x), tab (y), nrows*ncols );		\
  return Matrix (r, nrows, ncols);			\
}

#define MATRIX_SCALAR(name,op)			\
TMPL Matrix					\
operator op (const Matrix& x, const C c) {	\
  nat nrows= rows (x);                          \
  nat ncols= cols (x);                          \
  C* r= new C[nrows*ncols];			\
  name<C> (r, tab (x), c, nrows*ncols );	\
  return Matrix (r, nrows, ncols);		\
}

#define MATRIX_VAR_SCALAR(name,op)		\
TMPL Matrix					\
operator op (const C c, const Matrix& x) {	\
  nat nrows= rows (x);		      	        \
  nat ncols= cols (x);                          \
  C** r= new C[nrows*ncols];		       	\
  name<C> (r, tab (x), c, nrows*ncols);      	\
  return Matrix (r, nrows, ncols);		\
}

#define MATRIX_INLINE_BINARY(name,op)		\
TMPL Matrix&					\
operator op (Matrix& x, const Matrix& y) {	\
  nat nrows= rows (x);			        \
  assert (rows (y) == nrows);                   \
  nat ncols= cols (x);                          \
  assert (cols(y) == ncols);                    \
  name<C> (tab (x), tab (y), nrows*ncols);	\
  return x;					\
}  

#define MATRIX_INLINE_SCALAR(name,op)		\
TMPL Matrix&					\
operator op (Matrix& x, const C& c) {		\
  nat nrows= rows (x);			        \
  nat ncols= cols (x);                          \
  name<C> (tab (x), c, nrows*ncols);	     	\
  return x;					\
}  

#define MATRIX_TEST(name,op)				\
TMPL bool						\
operator op (const Matrix& x, const Matrix& y) {	\
  nat nrows= rows (x);					\
  assert (rows (y) == nrows );				\
  nat ncols= cols (x);					\
  assert (cols (y) == ncols);				\
  return name<C> (tab (x), tab (y), nrows*ncols);	\
}

MATRIX_UNARY(segment_neg,-);
MATRIX_BINARY(segment_add,+);
MATRIX_BINARY(segment_sub,-);
MATRIX_SCALAR(segment_mul,*);
MATRIX_VAR_SCALAR(segment_mul,*);
MATRIX_SCALAR(segment_div,/);
MATRIX_SCALAR(segment_mod,%);
MATRIX_INLINE_BINARY(segment_add,+=);
MATRIX_INLINE_BINARY(segment_sub,-=);
MATRIX_INLINE_SCALAR(segment_mul,*=);
MATRIX_INLINE_SCALAR(segment_div,/=);
MATRIX_INLINE_SCALAR(segment_mod,%=);
MATRIX_TEST(segment_eq,==);
MATRIX_TEST(segment_lesseq,<=);
MATRIX_TEST(segment_gtreq,>=);

ARITH_INT_SUGAR(TMPL,Matrix,C);
STRICT_COMPARE_SUGAR(TMPL,Matrix)

TMPL 
Matrix copy (const Matrix & x) {
  nat nrows=rows(x), ncols=cols(x);
  C* r = new C[nrows*ncols];
  block_copy (r, tab(x), nrows, ncols);
  return Matrix (r, ncols, nrows);
}

TMPL 
Matrix transpose (const Matrix & x) {
  nat nrows=rows(x), ncols=cols(x);
  C* r = new C[nrows*ncols];
  block_transpose (r, tab(x), nrows, ncols);
  return Matrix (r, ncols, nrows);
}

TMPL Matrix
submatrix (const Matrix& x, nat rdeb, nat rfin, nat cdeb, nat cfin) {
  nat nrows=rows(x), ncols=cols(x);
  C* r = new C[(rfin-rdeb+1)*(cfin-cdeb+1)];
  block_extraction (r, tab(x), nrows, rdeb, rfin, ncols, cdeb, cfin);
  return Matrix (r, rfin-rdeb, cfin-cdeb);
}

TMPL Matrix
operator * (const Matrix& x, const Matrix& y){
  nat xrows=rows(x), xcols=cols(x), ycols=cols(y);
  assert ( rows(y) == xcols );
  C* r = new C[xrows*ycols];
  block_naive_mul (r, tab(x), tab(y), xrows, xcols, ycols);
  return Matrix (r, xrows, ycols);
}

TMPL void
inline_swap_row (Matrix& m, nat i, nat j) {
  nat mrows=rows(m), mcols=cols(m);
  assert( (i < mrows) && (j < mrows) );
  block_inline_row_swap (tab(m), i, j, mrows, mcols);
}

TMPL void
inline_swap_col(Matrix& m, nat i, nat j) {
  nat mrows=rows(m), mcols=cols(m);
  assert( (i < mcols) && (j < mcols) );
  block_inline_col_swap (tab(m), i, j, mrows, mcols);
}

#define INLINE_SCAL_LINE_OP(name1,name2)	\
TMPL void name1 (Matrix& m, C c, nat i) {	\
  nat mrows=rows(m), mcols=cols(m);		\
  name2 (tab(m), c, i, mrows, mcols);		\
}

#define INLINE_LINES_OP(name1, name2)			\
TMPL void name1 (Matrix& m, nat i, C ci, nat j, C cj) {	\
  nat mrows=rows(m), mcols=cols(m);			\
  name2 (tab(m), i, ci, j, cj, mrows, mcols);		\
}

INLINE_SCAL_LINE_OP(inline_row_mul, block_inline_row_mul);
INLINE_SCAL_LINE_OP(inline_row_div, block_inline_row_div);
INLINE_SCAL_LINE_OP(inline_row_add, block_inline_row_add);
INLINE_SCAL_LINE_OP(inline_row_sub, block_inline_row_sub);
INLINE_SCAL_LINE_OP(inline_col_mul, block_inline_col_mul);
INLINE_SCAL_LINE_OP(inline_col_div, block_inline_col_div);
INLINE_SCAL_LINE_OP(inline_col_add, block_inline_col_add);
INLINE_SCAL_LINE_OP(inline_col_sub, block_inline_col_sub);
INLINE_LINES_OP(inline_rows_add,block_inline_rows_add);
INLINE_LINES_OP(inline_cols_add,block_inline_cols_add);
INLINE_LINES_OP(inline_rows_sub,block_inline_rows_sub);
INLINE_LINES_OP(inline_cols_sub,block_inline_cols_sub);

#undef TMPL
#undef Matrix
#undef Matrix_rep
} // namespace mml
#endif // __MML_DENSE_MATRIX_HPP
