/******************************************************************************
* MODULE     : dense_pivot.hpp
* DESCRIPTION: Pivoting for dense matrices
* COPYRIGHT  : (C) 2005  Olivier Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_DENSE_PIVOT_HPP
#define __MML_DENSE_PIVOT_HPP
#include "dense_matrix.hpp"


namespace mml {
#define TMPL template<typename C, typename M>
#define Matrix matrix<C,dense<M> >
#define Matrix_rep matrix_rep<C,dense<M> >


/** function to test if a pivot is admissible **/
template<class C>
inline bool IsAdmissible(const C& val) {
  if (val != C(0) ) 
    return true;
  else 
    return false;
}


/** function to test if a pivot is admissible **/
/** for numerical ground field matrix **/
template<class C>
inline bool IsAdmissible(const C& val, const C& tol) {
  if ( val > tol ) 
    return true;
  else 
    return false;
}


/** naive procedure to search an admissible pivot in a sub-matrix **/ 
TMPL void
SearchPivot( const Matrix& m, nat& ri, nat& ci, const nat& rf, 
	     const nat& cf,  bool& found=false) {
  assert ( (ri < rows(m)) && (ci < cols(m)) && ( rf < rows(m) ) 
	   && ( cf < cols(m) ) && (ri < rf) && (ci < cf) );
  nat i(ri), j(ci);
  C piv = m(i,j);
  for (; ( (! IsAdmissible(piv) ) && (i <= rf) ); i++) {
    for( j = ci; ( (! IsAdmissible(piv) ) && (j <=  cf) ); j++) {
      piv = m(i,j); 
    }
  }
  found = IsAdmissible(piv);
  if (found == true) {
    ci = j;
    ri = i;
  }
}


/** naive procedure to search an admissible pivot in a numerical **/
/** coefficient ground field matrix **/ 
TMPL void
SearchPivot( const Matrix& m, nat& ri, nat& ci, const nat& rf,
	     const nat& cf, bool& found=false, const C& tol) {
  assert ( (ri < rows(m)) && (ci < cols(m)) && ( rf < rows(m) ) && 
	   ( cf < cols(m) ) && ( ri < rf) && ( ci < cf) );
  nat i(ri), j(ci);
  C piv = m(i,j);
  for (; ( (! IsAdmissible(piv,tol) ) && (i < rf) ); i++) {
    for( j = ci; ( (! IsAdmissible(piv,tol) ) && (j <  cf) ); j++) {
      piv = m(i,j); 
    }
  }
  found = IsAdmissible(piv,tol);
  if (found == true) {
    ci = j;
    ri = i;
  }
}


  /* Brutal method */
TMPL void
gauss_pivoting(Matrix& m, const nat& ri, const nat& ci, const nat& rf, 
	      const nat& cf) {
  assert ( ( ri < rows(m) ) && ( ci < cols(m) ) && ( rf < rows(m) )
	   && ( cf < cols(m) ) );
  nat i(ri), j(ci);
  bool found = false;
  SearchPivot(m, i, j, rf, cf, found);
  //if we found a pivot
  if (found == true) {
    if ( i != ri ) 
      inline_swap_row(m,ri,i);
    if ( j != ci )
      inline_swap_col(m,ci,j);
    //pivot is now on the top and the left
    nat nb_index = min(rf-ri,cf-ci);
    //operations are made on the rows
    for ( nat index = 1; index <= nb_index; index++) 
      inline_rows_sub(m, ri + index, m(ri,ci), ri, m(ri+index,ci));
    //call the pivoting on the non triangular sub-matrix
    if ( (ri < rf) && (ci < cf) ) 
      gauss_pivoting(m, ri+1, ci+1, rf, cf);
  }
}


/* Bareiss method */
TMPL void
facto_gauss_pivoting (Matrix& m, const nat& ri, const nat& ci, const nat& rf, 
		      const nat& cf) {
  assert ( ( ri < rows(m) ) && ( ci < cols(m) ) && ( rf < rows(m) )
	   && ( cf < cols(m) ) );
  nat i(ri), j(ci);
  bool found = false;
  SearchPivot(m, i, j, rf, cf, found);
  //if we found a pivot
  if (found == true) {
    if ( i != ri ) 
      inline_swap_row(m,ri,i);
    if ( j != ci )
      inline_swap_col(m,ci,j);
    //pivot is now on the top and the left
    nat nb_index = min(rf-ri,cf-ci);
    //operations are made on the rows
    for ( nat index = 1; index <= nb_index; index++) {
      C p = gcd ( m(ri,ci) , m(ri+index,ci) );
      inline_rows_sub(m, ri + index, m(ri,ci)/p, ri, m(ri+index,ci)/p);
    }
    //call the pivoting on the non triangular sub-matrix
    if ( (ri < rf) && (ci < cf) ) 
      facto_gauss_pivoting(m, ri+1, ci+1, rf, cf);
  }
}


/* Brutal rank computation using diagonalisation */
TMPL nat
rank (const Matrix& m) {
  Matrix c(copy(m));
  gauss_pivoting(c,0,0,rows(c)-1,cols(c)-1);
  nat i(0);
  for (; ( ( c(i,i) != 0 ) && ( i < min( rows(c),cols(c) ) )) ; i++ ) {}
  return i;
}



/* Bareiss rank computation */
TMPL nat 
facto_rank (Matrix& m) {
  Matrix c(copy(m));
  facto_gauss_pivoting(c,0,0,rows(c)-1,cols(c)-1);
  nat i(0);
  for (; ( ( c(i,i) != 0 ) && ( i < min( rows(c), cols(c) ) ) ) ; i++) {}
  return i;
}


/* computation of determinant  by fraction free bareiss method */ 
TMPL C
facto_det (Matrix& m) {
  assert( cols(m) == rows(m) );
  Matrix c(copy(m));
  facto_gauss_pivoting(c,0,0,rows(c)-1, cols(c)-1);
  return( c(rows(c)-1,rows(c)-1) );
}

}//end of namespace

#endif //__MML_DENSE_PIVOT
