
/******************************************************************************
* MODULE     : real.hpp
* DESCRIPTION: Effective real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_REAL_HPP
#define __MML_REAL_HPP
#include "interval.hpp"
namespace mml {

#define MAX_CANCEL_TEST mml_default_zt

class real_rep REP_STRUCT {
public:
  const dyadic wt;  // weight of real number as an expression
  mp_exp_t magn;    // order of magnitude of the number
  interval best;    // the best currently available approximation
                    // FIXME: best should be protected
public:
  inline real_rep (const dyadic& wt2):
    wt (wt2), best (interval::fuzzy) {}
  inline real_rep (const dyadic& wt2, const interval& best2):
    wt (wt2), best (best2) {}
  virtual inline ~real_rep () {}
  inline void init_magn () { magn= expo (best); }
  inline void init () { approx (lp_plus_infinity); init_magn (); }

  virtual generic expression () const = 0;
  virtual void approx (const dyadic& err) = 0;
};

class real {
INDIRECT_PROTO (real, real_rep)
public:
  real (signed int i= 0);
  real (unsigned int i);
  real (signed short int i);
  real (unsigned short int i);
  real (signed long int i);
  real (unsigned long int i);

  real (const integer& x);
  real (const rational& x);

  static real log2 ();
  static real pi ();
  static real euler ();
  //static real catalan ();

  friend dyadic bound (const real& x);
  friend interval approx (const real& x, const dyadic& err);
  friend interval approx_linear (const real& x, const dyadic& err);
  friend interval approx_mult (const real& x, const dyadic& err);
  friend interval approximate (const real& x, const dyadic& err);
  friend interval expand (const real& x, mp_prec_t prec= mml_default_prec);
  friend bool normalize (const real& x, mp_prec_t max_cancel= MAX_CANCEL_TEST,
			 const dyadic& rel_err= error (1, -2));
  inline friend nat hash (const real& x) {
    interval i= expand (x, BITS_PER_LIMB+1);
    return hash (i);
  }
};
INDIRECT_IMPL (real, real_rep)

inline dyadic weight (const real& x1) {
  return x1->wt; }
inline dyadic weight (const real& x1, const real& x2) {
  return add_up (x1->wt, x2->wt); }

generic as_generic (const real& x);
inline ostream& operator << (ostream& out, const real& x) {
  return out << as_generic (x); }

real as_real (double x);
real as_real (const dyadic& x);

real operator - (const real& x);
real operator + (const real& x, const real& y);
real operator - (const real& x, const real& y);
real operator * (const real& x, const real& y);
real operator / (const real& x, const real& y);
real operator % (const real& x, const real& y);
real operator << (const real& x, mp_exp_t shift);
real operator >> (const real& x, mp_exp_t shift);
real sqrt (const real& x);

real abs (const real& x);
real min (const real& x, const real& y);
real max (const real& x, const real& y);
int sign (const real& x, int max_cancel= MAX_CANCEL_TEST);
integer floor (const real& x, int max_cancel= MAX_CANCEL_TEST);
integer trunc (const real& x, int max_cancel= MAX_CANCEL_TEST);
integer ceil  (const real& x, int max_cancel= MAX_CANCEL_TEST);
integer round (const real& x, int max_cancel= MAX_CANCEL_TEST);

real exp (const real& x);
real log (const real& x);
real cos (const real& x);
real sin (const real& x);
real asin (const real& x);
real acos (const real& x);
real atan (const real& x);
real atan2 (const real& y, const real& x);

ARITH_INT_SUGAR(,real,real);
SHIFT_INT_SUGAR(,real,real);
COMPARE_SUGAR(,real);

}
#endif // __MML_REAL_HPP
