
/******************************************************************************
* MODULE     : real.cpp
* DESCRIPTION: Effective real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "real.hpp"
#include "string.hpp"
namespace mml {

/******************************************************************************
* Different types of approximations
******************************************************************************/

dyadic
bound (const real& x) {
  if (is_signed (x.rep->best) || is_epsilon (x.rep->best))
    return bound (x.rep->best);
  else return dyadic (1, x.rep->magn, DEFAULT_LOW);
}

interval
approx (const real& x, const dyadic& err) {
  if (wider (x.rep->best, err)) {
    // cout << "."; cout.flush ();
    x.rep->approx (err);

    // A non zero exact result may lead to subtle errors, because its bit
    // precision is unknown (typically lost after normalization). Therefore,
    // we need to replace the exact result by a small surrounding interval
    // of radius < err.
    if (is_non_zero_exact (x.rep->best) && expo (err) < expo (x.rep->best)) {
      mp_prec_t p= (mp_prec_t) 2 + expo (x.rep->best) - expo (err);
      x.rep->best= extend (x.rep->best, p);
    }

    // The precision of a computed approximation can be a little bit too low
    // due to rounding errors and possible switching to low precision
    // representation when normalizing high precision intervals.
    // We slightly increase the precision if this problem occurs.
    // If the computed precision is far too low, then this probably
    // indicates a bug in the code.
    if (wider (x.rep->best, err)) {
      dyadic factor= radius (x.rep->best) / err;
      if (factor > 16) {
	cerr << "\nPrecision deficiency: " << factor << "\n";
	cerr << "rad= " << radius (x.rep->best) << "\n";
	cerr << "err= " << err << "\n";
	cerr << "x  = " /*<< set_real_expr (true)*/ << x->expression() << "\n";
	assert (factor <= 16);
      }
      return approx (x, err / (factor + 1));
    }
  }
  return x.rep->best;
}

interval
approx_linear (const real& x, const dyadic& err) {
  // Approximate to precision err. If a larger precision is obtained,
  // then truncate to at most twice the required precision, so as to
  // make linear time algorithms based on the result as efficient as possible.
  interval r= approx (x, err);
  mp_prec_t given= prec (r), required;
  if (is_infinity (err)) required= 0;
  else required= max (expo (r) - expo (err), (mp_exp_t) 0);
  required <<= 1;
  if (required >= given) return r;
  if (required < DEFAULT_LOW) required= DEFAULT_LOW;
  return reduce (r, required);
}

interval
approx_mult (const real& x, const dyadic& err) {
  // Approximate to precision err. If a larger precision is obtained,
  // then truncate to a precision which make O(M(n)) algorithms which
  // use the result as efficient as possible
  interval r= approx (x, err);
  mp_prec_t given= prec (r), required;
  if (is_infinity (err)) required= 0;
  else required= max (expo (r) - expo (err), (mp_exp_t) 0);
  if (required < (BITS_PER_LIMB << 8)) required += (required >> 1);
  else required <<= 1;
  if (required >= given) return r;
  if (required < DEFAULT_LOW) required= DEFAULT_LOW;
  return reduce (r, required);
}

interval
approximate (const real& x, const dyadic& err) {
  // Approximate to precision err. Truncate if a larger precision is obtained.
  interval r= approx (x, err);
  if (is_epsilon (r)) return r;
  mp_exp_t ret_expo= expo (r);
  mp_exp_t err_expo= expo (err);
  return truncate (r, ret_expo - err_expo + 5);
  // NOTE: add 4+1 to precision to keep last decimal digit when printing
}

/******************************************************************************
* Normalization of non-zero real numbers
******************************************************************************/

bool
normalize (const real& x, mp_prec_t max_cancel, const dyadic& rel_err) {
  if (is_infinity (x.rep->best)) return false;
  interval r= approx (x, lp_plus_infinity);
  dyadic delta= bound (x) >> max_cancel;
  while (true) {
    if (radius (r) < rel_err * abs (center (r))) return false;
    if (delta == 0 || is_epsilon (r) || abs (radius (r)) < delta) {
      if (max_cancel >= MAX_CANCEL_TEST) {
	real_rep* rep= (real_rep*) ((void*) x.rep);
	rep->best= dyadic (0);
      }
      return true;
    }
    r= approx (x, radius (r) >> BITS_PER_LIMB);
  }
}

interval
expand (const real& x, mp_prec_t prec) {
  if (normalize (x, MAX_CANCEL_TEST, error (1, -(prec+1)))) return 0;
  return truncate (x.rep->best, prec);
}

int
sign (const real& x, int max_cancel) {
  if (normalize (x, max_cancel)) {
    /*cerr << "Warning: heuristic determination of sign of " << x << "\n"*/;
  }
  if (x->best > 0) return 1;
  if (x->best < 0) return -1;
  return 0;
}

/******************************************************************************
* Conversions
******************************************************************************/

generic
as_generic (const real& x) {
  if (mml_default_expr) return x->expression ();
  else if (x == 0) return 0;
  else return print_to_string (expand (x, 20));
}

}
