
/******************************************************************************
* MODULE     : real_div.cpp
* DESCRIPTION: Division of real numbers
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "real.hpp"
namespace mml {

class div_real_rep: public real_rep {
private:
  real x, y;
public:
  inline div_real_rep (const real& x2, const real& y2):
    real_rep (weight (x2, y2)), x (x2), y (y2) { init (); }
  generic expression () const {
    return over (as_generic (x), as_generic (y)); }
  void approx (const dyadic& err);
};

void
div_real_rep::approx (const dyadic& err) {
  if (has_huge_prec (best)) {
    dyadic rel_r= err / magnitude (best);
    dyadic rel_x= spread (rel_r, x->wt, wt);
    dyadic rel_y= spread (rel_r, y->wt, wt);
    best= approx_mult (x, magnitude (x->best) * rel_x) /
          approx_mult (y, magnitude (y->best) * rel_y);
  }
  else {
    interval xa= mml::approx (x, lp_plus_infinity);
    if (is_epsilon (xa)) { best= xa; return; }
    dyadic bound_x= bound (x);
    dyadic bound_y= bound (y);
    dyadic bound_r= bound_x / bound_y;
    dyadic rel_r  = min (err / bound_r, 1);
    if ((rel_r << (DEFAULT_LOW<<1)) < 1) {
      // Since we lack information for making good a priori error estimates,
      // we first compute an approximation at a more modest precision and
      // then check whether we may use the huge precision error estimates.
      dyadic modest_err= sqrt (rel_r) * bound_r;
      approx (modest_err);
      if (has_huge_prec (best)) { approx (err); return; }
    }
    dyadic rel_x= spread (rel_r, x->wt, wt);
    dyadic rel_y= spread (rel_r, y->wt, wt);
    best= approx_mult (x, rel_x * bound_x) /
          approx_mult (y, rel_y * bound_y);
  }
}

real::real (const rational& x): rep (new div_real_rep (num (x), den (x))) {}

real
operator / (const real& x, const real& y) {
  if (normalize (y, MAX_CANCEL_TEST, error (1))) return 0;
  return real (new div_real_rep (x, y));
}

real
operator % (const real& x, const real& y) {
  if (normalize (y, MAX_CANCEL_TEST, error (1))) return x;
  return 0;
}

} // namespace mml
