
/******************************************************************************
* MODULE     : block.hpp
* DESCRIPTION: header for low level matricial computations
* COPYRIGHT  : (C) 2004  Olivier Ruatta
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__BLOCK__HPP
#define __MML__BLOCK__HPP
#include "segment.hpp"

namespace mml {
#define TMPL template<class C>

TMPL void
block_set (C* dest, const C& c, nat rows, nat cols) {
  segment_clear (dest, rows*cols);
  nat m= min (rows, cols);
  while (m != 0) {
    *dest= c;
    dest += cols+1; m--;
  }
}

TMPL inline void
block_copy (C* dest, const C* s, nat rows, nat cols) {
  segment_copy (dest, s, rows*cols);
}


TMPL void
block_extraction (C* dest, const C* s,
		  nat rows, nat nrd, nat nrf,
		  nat cols, nat ncd, nat ncf)
{
  const C* is = s;
  C* id = dest;
  is += nrd*cols;
  for(nat i = 0; i < nrf - nrd + 1; i++) {
    is += ncd;
    for(nat j = 0; j < ncf - ncd + 1; j++) {
      *id = *is;
      id++; is++;
    }
    is += cols - ncf - 1;
  } 
}


TMPL void
block_duplicate (C* dest, nat dr, nat dc, nat ddr, nat ddc,
		 C* b, nat br, nat bc, nat dbr, nat fbr, nat dbc, nat fbc) {
  nat sr = fbr - dbr + 1;
  nat sc = fbc - dbc + 1;
  assert( (dbr < fbc ) && (fbc < bc) );
  assert( (ddr + sr - 1 < dr) && (dbc + sc - 1 < dc) );
  C* id = dest + ddr * dc;
  C* ib = b + dbr * bc;
  for(nat i = 0; i < sr; i++) {
    id += ddc;
    ib += dbc;
    for(nat j = 0; j < sc; j++) {
      *id = *ib;
      id++; ib++;
    }
    id += dc - (ddc + sc);
    ib += bc - fbc - 1;
  }
}

TMPL void
block_transpose (C* dest, const C* s, nat rows, nat cols) {
  nat trows;
  nat tcols(0);
  while (tcols != cols) {
    trows=0;
    while (trows != rows) {
      *dest = *(s + trows*cols);
      dest++; trows++;
    }
    tcols++; s++;
  }
}

TMPL void
block_naive_mul (C* dest, const C* s1, const C* s2, nat r1, nat c1, nat c2) {
  C tmp;
  nat i, j, k;
  for (i=0; i<r1; i++) {
    for (j=0; j<c2; j++) {
      tmp= 0;
      for (k=0; k<c1; k++) {
	tmp += s1[i*c1+k] * s2[j+k*c2];
      }
      dest[i*c2+j]= tmp;
    }
  }
}


#define BLOCK_INLINE_SCAL_ROW_OP(name,op)			\
TMPL void name(C* s, const C& sc, nat i, nat r, nat c) {	\
  assert (i<r);							\
  C* it= s+i*c;							\
  while (c>0) {							\
    *it op sc;							\
    it++; c--;							\
  }								\
}

#define BLOCK_INLINE_SCAL_COL_OP(name,op)			\
TMPL void name (C* s, const C& sc, nat i, nat r, nat c) {	\
  assert (i<c);							\
  C* it= s+i;							\
  while (r>0) {							\
    *it op sc;							\
    it+=c; r--;							\
  }								\
}

#define BLOCK_INLINE_ROWS_OP(name,op)					      \
TMPL void name (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) { \
  assert ((i<r) && (j<r));						      \
  C* iti= s + i*c;							      \
  C* itj= s + j*c;							      \
  while (c>0){								      \
    *iti = si * (*iti) op sj * (*itj);					      \
    iti++; itj++; c--;							      \
  }									      \
}

#define BLOCK_INLINE_COLS_OP(name,op)					      \
TMPL void name (C* s, nat i, const C& si, nat j, const C& sj, nat r, nat c) { \
  assert ((i<c) && (j<c));						      \
  C* iti= s+i, itj= s+j;						      \
  while (r>0) {								      \
    *iti = si*(*iti) op  sj*(*itj);					      \
    iti+=c; itj+=c; r--;						      \
  }									      \
}

#define BLOCK_OP(name,op)                                                     \
TMPL void name (C* dest, nat dr, nat dc, nat ddr, nat fdr, nat ddc, nat fdc,  \
C* b1, nat br1, nat bc1, nat dbr1, nat fbr1, nat dbc1, nat fbc1,              \
C* b2, nat br2, nat bc2, nat dbr2, nat fbr2, nat dbc2, nat fbc2) {            \
  assert( (ddr < fdr) && (fdr < dr) && (ddc < fdc) && (fdc < dc) );           \
  assert( (dbr1 < fbr1) && (fbr1 < br1) && (dbc1 < fbc1) && (fbc1 < bc1) );   \
  assert( (dbr2 < fbr2) && (fbr2 < br2) && (dbc2 < fbc2) && (fbc2 < bc2) );   \
  assert( (fdr - ddr == fbr1 - dbr1) && (fbr1 - dbr1 == fbr2 - dbr2) );       \
  assert( (fdc - ddc == fbc1 - dbc1) && (fbc1 - dbc1 == fbc2 - dbc2) );       \
  nat sr = fdr - ddr + 1;                                                     \
  nat sc = fdc - ddc + 1;                                                     \
  C* id = dest + ddr * dc;                                                    \
  C* ib1 = b1 + dbr1 * bc1;                                                   \
  C* ib2 = b2 + dbr2 * bc2;                                                   \
  for (nat i = 0; i < sr; i++) {                                              \
    id += ddc;                                                                \
    ib1 += dbc1;                                                              \
    ib2 += dbc2;                                                              \
    for (nat j = 0; j < sc; j++) {                                            \
      *id = *ib1 op *ib2;                                                     \
      id++; ib1++; ib2++;                                                     \
    }                                                                         \
    id += dc - (ddc + sc);                                                    \
    ib1 += bc1 - fbc1 - 1;                                                    \
    ib2 += bc2 - fbc2 - 1;                                                    \
  }                                                                           \
}
		

#define BLOCK_INLINE_OP(name, op)                                             \
TMPL void name (C* b, nat br, nat bc, nat dbr, nat fbr, nat dbc, nat fbc,     \
		C* d, nat dr, nat dc, nat ddr, nat fdr, nat ddc, nat fdc) {   \
  assert( (dbr < fbr) && (fbr < br) && (dbc < fbc) && (fbc < bc) );           \
  assert( (ddr < fdr) && (fdr < dr) && (ddc < fdc) && (fdc < dc) );           \
  assert( (fbr - dbr == fdr - ddr) && (fbc - dbc == fdc - ddc) );             \
  nat sr = fbr - dbr + 1;                                                     \
  nat sc = fbc - dbc + 1;                                                     \
  C* ib = b + dbr * bc;                                                       \
  C* id = d + ddr * dc;                                                       \
  for (nat i = 0; i < sr; i++) {                                              \
    ib += dbc;                                                                \
    id += ddc;                                                                \
    for(nat j = 0; j < sc; j++) {                                             \
      *ib op *id;                                                             \
      ib++; id++;                                                             \
    }                                                                         \
    ib += bc - fbc - 1;                                                       \
    id += dc - fdc - 1;                                                       \
  }                                                                           \
}

BLOCK_INLINE_SCAL_ROW_OP(block_inline_row_mul,*=);
BLOCK_INLINE_SCAL_ROW_OP(block_inline_row_div,/=);
BLOCK_INLINE_SCAL_ROW_OP(block_inline_row_add,+=);
BLOCK_INLINE_SCAL_ROW_OP(block_inline_row_sub,-=);
BLOCK_INLINE_SCAL_COL_OP(block_inline_col_mul,*=);
BLOCK_INLINE_SCAL_COL_OP(block_inline_col_div,/=);
BLOCK_INLINE_SCAL_COL_OP(block_inline_col_add,+=);
BLOCK_INLINE_SCAL_COL_OP(block_inline_col_sub,-=);
BLOCK_INLINE_ROWS_OP(block_inline_rows_add,+);
BLOCK_INLINE_ROWS_OP(block_inline_rows_sub,-);
BLOCK_INLINE_COLS_OP(block_inline_cols_add,+);
BLOCK_INLINE_COLS_OP(block_inline_cols_sub,-);
BLOCK_OP(block_add,+);
BLOCK_OP(block_sub,-);
BLOCK_INLINE_OP(block_inline_add,+=);
BLOCK_INLINE_OP(block_inline_sub,-=);


TMPL void
block_inline_row_swap (C* s, nat i, nat j, nat r, nat c) {
  assert ((i<r) && (j<r));
  C* ri= s + i*c;
  C* rj= s + j*c;
  C piv;
  while (c > 0) { 
    piv= *ri; *ri= *rj; *rj= piv;
    ri++; rj++; c--;
  }
}

TMPL void
block_inline_col_swap (C* s, nat i, nat j, nat r, nat c) {
  assert((i<c) && (j<c));
  C* ci= s+i;
  C* cj= s+j;
  C piv;
  while (r>0) {
    piv= *ci; *ci= *cj; *cj= piv;
    ci+=c; cj+=c; r--;
  }
}

#undef TMPL
}//namespace mml
#endif //__MML__BLOCK__HPP
