
/******************************************************************************
* MODULE     : karatsuba.hpp
* DESCRIPTION: low level pair/odd Karatsuba multiplication
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML__KARATSUBA__HPP
#define __MML__KARATSUBA__HPP
#include "jet.hpp"

namespace mml {
#define TMPL template<class C>

/******************************************************************************
* Vectorial dichotomy operations
******************************************************************************/

TMPL inline void
jet_half (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest= *s;
    dest++; s += 2; l--;
  }
}

TMPL inline void
jet_double (C* dest, const C* s, nat l) {
  while (l != 0) {
    *dest= *s;
    dest += 2; s++; l--;
  }
}

/******************************************************************************
* Multiplication
******************************************************************************/

template<typename C, nat Bound>
class multiplier<C,karatsuba<Bound> > {
  static void multiply (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n2 < Bound) { jet_mul<C,naive> (dest, s1, s2, n1, n2); return; }
    nat p1= n1 >> 1, p2= n2 >> 1, P= p1+p2-1, spc= 3*(p1+p2);
    INLINE_ARRAY (C, buf, spc);
    C* low1= dest;
    C* low2= low1 + p1;
    C* mid1= buf;
    C* mid2= buf + p1;
    C* hi1 = mid2 + p2;
    C* hi2 = hi1 + p1;
    C* Low = mid1;
    C* Mid = hi1;
    C* Hi  = Mid + p1 + p2;
    jet_half (low1, s1, p1);
    jet_half (hi1 , s1+1, p1);
    jet_add  (mid1, low1, hi1, p1);
    jet_half (low2, s2, p2);
    jet_half (hi2 , s2+1, p2);
    jet_add  (mid2, low2, hi2, p2);
    multiply (Hi , hi1 , hi2 , p1, p2);
    multiply (Mid, mid1, mid2, p1, p2);
    multiply (Low, low1, low2, p1, p2);
    jet_sub  (Mid, Low, P);
    jet_sub  (Mid, Hi , P);
    jet_add  (Low+1, Hi, P-1);
    Low[P]= Hi[P-1];
    jet_double (dest, Low, P+1);
    jet_double (dest+1, Mid, P);
    DELETE_ARRAY (C, buf, spc);

    if ((n1 & 1) != 0) {
      dest[(P<<1)+1]= C(0);
      jet_mul_add (dest + (p1<<1), s2, s1[p1<<1], p2<<1);
    }
    if ((n2 & 1) != 0) {
      dest[n1+n2-2]= C(0);
      jet_mul_add (dest + (p2<<1), s1, s2[p2<<1], n1);
    }
  }

public:
  static void mul (C* dest, const C* s1, const C* s2, nat n1, nat n2) {
    if (n1 == 0 || n2 == 0) return;
    if (n1 < n2) multiply (dest, s2, s1, n2, n1);
    else multiply (dest, s1, s2, n1, n2);
  }
};

/******************************************************************************
* Division
******************************************************************************/

template<typename C, nat Bound>
class divisor<C,karatsuba<Bound> > {
public:
  static void div_mod (C* dest, C* s1, const C* s2, nat n1, nat n2) {
    return
      jet_dicho_div_mod<C,karatsuba<Bound> > (dest, s1, s2, n1, n2, Bound);
  }
};

#undef TMPL
} // namespace mml
#endif //__MML__KARATSUBA__HPP
