
/******************************************************************************
* MODULE     : implicit_series.hpp
* DESCRIPTION: Implicit series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_IMPLICIT_SERIES_HPP
#define __MML_IMPLICIT_SERIES_HPP
#include "dense_vector.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename V=naive>
#define TMPL template<typename C, typename V>
#define Series series<C,V>
#define Series_rep series_rep<C,V>

/******************************************************************************
* Implicit series
******************************************************************************/

TMPL_DEF
class implicit_series_rep: public Series_rep {
  Series eq;
public:
  implicit_series_rep () {}
  C next () { return eq[this->n]; }
  generic expression () const { return generic ("implicit"); }
  C& initial (nat n2) {
    if (n2>=this->n) { this->n = n2+1; set_precision (this->n); }
    return this->a[n2]; }
  Series& equation () { return eq; }
};

TMPL Series
implicit_series () {
  return Series (new implicit_series_rep<C,V> ());
}

TMPL_DEF
class build_series_rep: public Series_rep {
  Series f;              // the series
  string op;             // obtained using this operation
  vector<Series > args;  // on these arguments
public:
  build_series_rep (const Series& f2, const Series& eq,
		    const string& op2, const vector<Series >& args2):
    f (f2), op (op2), args (args2) { f.equation ()= eq; }
  ~build_series_rep () { f.equation ()= 0; }
  C next () { return f[this->n]; }
  generic expression () const {
    if (dim (args) == 0) return generic (op);
    generic ops= comma ();
    for (nat i=0; i<dim(args); i++)
      ops= comma (ops, as_generic (args[i]));
    return apply (op, ops);
  }
};

TMPL Series
build (const Series& f, const Series& eq,
       const string& op= "implicit") {
  vector<Series > v;
  return Series (new build_series_rep<C,V> (f, eq, op, v));
}

TMPL Series
build (const Series& f, const Series& eq,
       const string& op, const Series& a1) {
  return Series (new build_series_rep<C,V> (f, eq, op, vec (a1)));
}

TMPL Series
build (const Series& f, const Series& eq,
       const string& op, const Series& a1, const Series& a2) {
  return Series (new build_series_rep<C,V> (f, eq, op, vec (a1, a2)));
}

/******************************************************************************
* Implicit systems
******************************************************************************/

TMPL vector<Series >
implicit_vector_series (nat n) {
  Series dummy;
  vector<Series > v (dummy, n);
  for (nat i=0; i<n; i++)
    v(i)= implicit_series<C,V> ();
  return v;
}

TMPL_DEF
class system_series_rep: public Series_rep {
  Series f;              // a series depending on the implicit system
  vector<Series > v;     // the vector of implicit series
  string op;             // obtained using this operation
  vector<Series > args;  // on these arguments
public:
  system_series_rep (const Series& f2,
		     const vector<Series >& v2, const vector<Series >& eq,
		     const string& op2, const vector<Series >& args2):
    f (f2), v (v2), op (op2), args (args2)
  {
    for (nat i=0; i<dim(v); i++)
      v(i).equation ()= eq[i];
  }
  ~system_series_rep () {
    for (nat i=0; i<dim(v); i++)
      v(i).equation ()= 0;
  }
  C next () { return f[this->n]; }
  generic expression () const {
    if (dim (args) == 0) return generic (op);
    generic ops= comma ();
    for (nat i=0; i<dim(args); i++)
      ops= comma (ops, as_generic (args[i]));
    return apply (op, ops);
  }
};

TMPL Series
build (const Series& f,
       const vector<Series >& v, const vector<Series >& eq,
       const string& op= "implicit") {
  vector<Series > z;
  return Series (new system_series_rep<C,V> (f, v, eq, op, z));
}

TMPL Series
build (const Series& f,
       const vector<Series >& v, const vector<Series >& eq,
       const string& op, const Series& a1) {
  return Series (new system_series_rep<C,V> (f, v, eq, op, vec (a1)));
}

/******************************************************************************
* Division, square root, exponentiation and logarithm
******************************************************************************/

TMPL Series
operator / (const Series& f, const Series& g) {
  if (g[0] == 0) {
    nat val;
    for (val=0; g[val]==0; val++)
      if (val == the_series_zt) return 0;
    return (f>>val) / (g>>val);
  }
  Series h= implicit_series<C,V> ();
  return build (h, (f - ((h * (g>>1)) << 1)) / g[0], "/", f, g);
}

TMPL Series
operator % (const Series& f, const Series& g) {
  return f - g * (f / g);
}

TMPL Series
sqrt (const Series& f) {
  if (f[0] == 0) {
    nat val= 0;
    for (val=0; val<the_series_zt && f[val]==0; val++);
    assert (val<the_series_zt && (val&1) == 0);
    return sqrt (f >> val) << (val/2);
  }
  Series g= implicit_series<C,V> ();
  g(0)= sqrt (f[0]);
  return build (g, integrate (derive (f) / (2*g)), "sqrt", f);
}

TMPL Series
pow (const Series& f, const C& c) {
  Series g= implicit_series<C,V> ();
  g(0)= pow (f[0], c);
  return build (g, integrate (c * derive (f) * g / f), "^", f, Series (c));
}

TMPL Series
exp (const Series& f) {
  Series g= implicit_series<C,V> ();
  g(0)= exp (f[0]);
  return build (g, integrate (g * derive (f)), "exp", f);
}

TMPL Series
log (const Series& f) {
  Series g= implicit_series<C,V> ();
  g(0)= log (f[0]);
  return build (g, integrate (derive (f) / f), "log", f);
}

/******************************************************************************
* Trigonometric functions
******************************************************************************/

TMPL Series
sin (const Series& f) {
  Series df= derive (f);
  vector<Series > sys= implicit_vector_series<C,V> (2);
  sys(0)(0)= sin (f[0]);
  sys(1)(0)= cos (f[0]);
  vector<Series > eq=
    vec (integrate (df * sys[1]), integrate (-df * sys[0]));
  return build (sys[0], sys, eq, "sin", f);
}

TMPL Series
cos (const Series& f) {
  Series df= derive (f);
  vector<Series > sys= implicit_vector_series<C,V> (2);
  sys(0)(0)= sin (f[0]);
  sys(1)(0)= cos (f[0]);
  vector<Series > eq=
    vec (integrate (df * sys[1]), integrate (-df * sys[0]));
  return build (sys[1], sys, eq, "cos", f);
}

TMPL Series
tan (const Series& f) {
  Series df= derive (f);
  vector<Series > sys= implicit_vector_series<C,V> (2);
  sys(0)(0)= sin (f[0]);
  sys(1)(0)= cos (f[0]);
  vector<Series > eq=
    vec (integrate (df * sys[1]), integrate (-df * sys[0]));
  return build (sys[0]/sys[1], sys, eq, "tan", f);
}

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
} // namespace mml
#endif // __MML_IMPLICIT_SERIES_HPP
