
/******************************************************************************
* MODULE     : generic_functions.cpp
* DESCRIPTION: Standard functions on generic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "generic_internal.hpp"
namespace mml {

/******************************************************************************
* N-ary operations
******************************************************************************/

generic
comma () {
  return generic (MML_COMMA) ();
}

generic
comma (const generic& g1, const generic& g2) {
  return generic (MML_COMMA) (g1, g2);
}

void
apply_sub (vector<generic>& v, const generic& g) {
  if (is_compound (g) && N(g) >= 1 && g[0] == MML_COMMA) {
    int i, n= N(g);
    for (i=1; i<n; i++)
      apply_sub (v, g[i]);
  }
  else v << g;
}

generic
apply (const generic& f, const generic& args) {
  vector<generic> v;
  v << f;
  apply_sub (v, args);
  return generic_compound (v);
}

generic
tuple (const generic& g) {
  return apply (MML_TUPLE, g);
}

generic
vtuple (const generic& g) {
  return apply (MML_VTUPLE, g);
}

generic
sqtuple (const generic& g) {
  return apply (MML_SQTUPLE, g);
}

/******************************************************************************
* Basic arithmetic
******************************************************************************/

generic
operator - (const generic& g1) {
  return generic (MML_MINUS) (g1);
}

generic
operator + (const generic& g1, const generic& g2) {
  return generic (MML_PLUS) (g1, g2);
}

generic
operator - (const generic& g1, const generic& g2) {
  return generic (MML_MINUS) (g1, g2);
}

generic
operator * (const generic& g1, const generic& g2) {
  return generic (MML_TIMES) (g1, g2);
}

generic
operator / (const generic& g1, const generic& g2) {
  return generic (MML_OVER) (g1, g2);
}

generic
operator % (const generic& g1, const generic& g2) {
  return generic (MML_MOD) (g1, g2);
}

generic
pow (const generic& g1, const generic& g2) {
  return generic (MML_POWER) (g1, g2);
}

/******************************************************************************
* Additive arithmetic with simplifications
******************************************************************************/

static void
signed_decompose (const generic& g, generic& abs_g, int& sgn_g) {
  if (is_atom (g)) {
    string s= as_string (g);
    if (N(s) > 1 && s[0] == '-' &&
	(s[1] == '.' || (s[1] >= '0' && s[1] <= '9'))) {
      abs_g= generic (s (1, N(s)));
      sgn_g= -1;
    }
    else {
      abs_g= g;
      sgn_g= 1;
    }
  }
  else if (is_func (g, MML_MINUS, 1)) {
    signed_decompose (g[1], abs_g, sgn_g);
    sgn_g= -sgn_g;
  }
  else if (is_func (g, MML_TIMES, 2) || is_func (g, MML_OVER, 2)) {
    generic a1, a2;
    int     s1, s2;
    signed_decompose (g[1], a1, s1);
    signed_decompose (g[2], a2, s2);
    if (is_func (g, MML_TIMES, 2)) abs_g= a1 * a2;
    else abs_g= a1 / a2;
    sgn_g= s1 * s2;
  }
  else {
    abs_g= g;
    sgn_g= 1;
  }
}

generic
negate (const generic& g) {
  if (g == 0) return g;
  if (is_func (g, MML_MINUS, 1))
    return g[1];
  if (is_func (g, MML_PLUS, 2))
    return minus (negate (g[1]), g[2]);
  if (is_func (g, MML_MINUS, 2))
    return plus (negate (g[1]), g[2]);
  generic abs_g; int sgn_g;
  signed_decompose (g, abs_g, sgn_g);
  if (sgn_g > 0) return -abs_g;
  else return abs_g;
}

generic
plus (const generic& g1, const generic& g2) {
  if (g1 == 0) return g2;
  if (g2 == 0) return g1;
  if (is_func (g2, MML_MINUS, 1))
    return minus (g1, g2[1]);
  if (is_func (g2, MML_PLUS, 2))
    return plus (plus (g1, g2[1]), g2[2]);
  if (is_func (g2, MML_MINUS, 2))
    return minus (plus (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (sgn_g2 < 0) return g1 - abs_g2;
  else return g1 + abs_g2;
}

generic
minus (const generic& g1, const generic& g2) {
  if (g2 == 0) return g1;
  if (is_func (g2, MML_MINUS, 1))
    return plus (g1, g2[1]);
  if (is_func (g2, MML_PLUS, 2))
    return minus (minus (g1, g2[1]), g2[2]);
  if (is_func (g2, MML_MINUS, 2))
    return plus (minus (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (g1 == 0) {
    if (sgn_g2 > 0) return - abs_g2;
    else return abs_g2;    
  }
  else {
    if (sgn_g2 > 0) return g1 - abs_g2;
    else return g1 + abs_g2;
  }
}

static generic
migrate_sub (const generic& g, const generic& h) {
  if (is_func (g, MML_MINUS, 2))
    return migrate_sub (g[1], h) - g[2];
  generic abs_g; int sgn_g;
  signed_decompose (g, abs_g, sgn_g);
  return h - abs_g;
}

static generic
migrate_sub (const generic& g, nat found) {
  if (found == 0) return migrate_sub (g[1], g[2]);
  return (g[0]) (migrate_sub (g[1], found-1), g[2]);
}

generic
migrate_negate (const generic& g, nat maximal) {
  int i= 0, found= -1;
  generic h= g;
  while (is_func (h, MML_PLUS, 2) || is_func (h, MML_MINUS, 2)) {
    if (is_func (h, MML_PLUS, 2)) found= i;
    if (maximal == 0) return g;
    h= h[1];
    maximal--;
    i++;
  }
  if (found == -1) return g;
  generic abs_h; int sgn_h;
  signed_decompose (h, abs_h, sgn_h);
  if (sgn_h >= 0) return g;
  return migrate_sub (g, found);
}

/******************************************************************************
* Multiplicative arithmetic with simplifications
******************************************************************************/

generic
times (const generic& g1, const generic& g2, bool frac_flag) {
  if (frac_flag && is_func (g1, MML_OVER, 2)) {
    if (is_func (g2, MML_OVER, 2))
      return times (g1[1], g2[1], frac_flag) / times (g1[2], g2[2], frac_flag);
    else return times (g1[1], g2, frac_flag) / g1[2];
  }
  if (frac_flag && is_func (g2, MML_OVER, 2))
    return times (g1, g2[1], frac_flag) / g2[2];
  if (g1 == 0) return 0;
  if (g2 == 0) return 0;
  if (g1 == 1) return g2;
  if (g2 == 1) return g1;
  if (g1 == -1) return negate (g2);
  if (g2 == -1) return negate (g1);
  return g1 * g2;
}

generic
over (const generic& g1, const generic& g2, bool frac_flag) {
  if (frac_flag && is_func (g2, MML_OVER, 2))
    return times (g1, g2[2] / g2[1], frac_flag);
  if (g2 == 1) return g1;
  if (g2 == -1) return negate (g1);
  if (frac_flag) return times (g1, 1 / g2, frac_flag);
  else return g1/g2;
}

generic
power (const generic& g1, const generic& g2, bool frac_flag) {
  if (g1 == 0) return 0;
  if (g1 == 1) return 1;
  if (g2 == 0) return 1;
  if (g2 == 1) return g1;
  if (g2 == generic (1) / generic (2))
    return generic ("sqrt") (g1);
  if (frac_flag) {
    generic abs; int sgn;
    signed_decompose (g2, abs, sgn);
    if (sgn < 0) return 1 / power (g1, abs);
  }
  return pow (g1, g2);
}

} // namespace mml
